# Copyright 2012 by the University of Oxford
# Copyright 2015 by Medical Research Council
# Author: Paul Emsley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA

from __future__ import print_function
from rdkit import Chem

# match_atom_index can be of type int or a list - otherwise trouble.
#
# Note that atom_type properties can also have been set in hydrogen_transformations():
#
def set_atom_type(match, match_atom_index, mol, atom_type, types_type='Refmac'):

    # Refmac and monomer_library mean the same thing.

    prop_str = 'type_energy'
    if types_type == 'Parm@Frosst':
        prop_str = 'pf_atom_type'
    try:
        # print('match {} match_atom_index {} types_type {} '.format(match, match_atom_index, types_type))
        this_atom_idx = match[match_atom_index]
        try:
            current_type = mol.GetAtomWithIdx(this_atom_idx).GetProp(prop_str)
        except KeyError:
            try:
                mol.GetAtomWithIdx(this_atom_idx).SetProp(prop_str, atom_type)
                name = mol.GetAtomWithIdx(this_atom_idx).GetProp("name")
                if False:
                    print('   actually have set atom number %s having name %s with type %s ' %
                          (str(this_atom_idx).rjust(2), repr(name), repr(atom_type)))
            except KeyError as e:
                if types_type != "Parm@Frosst":
                   print('No name for atom idx', match_atom_index)

    except TypeError:
        for match_atom in match_atom_index:
            # print('   set_atom_type ', match, match_atom, mol, atom_type, types_type)
            set_atom_type(match, match_atom, mol, atom_type, types_type)

def ele_to_smart(v):
    return (v.upper(), '['+v+']', 0)

# those not handled by hand-coding
def smarts_by_element():
   eles = [
      "He", "Li", "Be", "B",  "Ne", "Na", "Mg", "Al",
      "Ar", "K", "Ca",  "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu",
      "Zn", "Ga", "Ge", "As", "Se",      "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc",
      "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs", "Ba", "La",
      "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu",
      "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At",
      "Rn", "Fr", "Ra", "Ac", "Th", "Pa", "U"]
   return map(ele_to_smart, eles)

def set_monomer_library_atom_types(mol):
    smarts_list = [

        # Full coverage for C, H, O.

        # Oxygen
        ('O2',  "[OX2;H0]", 0), # ester, Os between P and C are O2, not OP
        ('OP',  'O~P',   0),
        ('OS',  'O~S',   0),
        ('OB',  'O~B',   0),
        ('OC',  '*C(=O)[OH]', (2,3)), # carboxylic acid
        ('OC',  '*C(=O)O',    (2,3)), # carboxylate, doesn't match deloc bonds
        ('OH1', '[OH1]', 0), # alcohol
        ('O2',  "[oX2;H0]", 0), # ring oxygen
        ('O',   'O=*',   0), # carbonyl oxygen

        # OH2 no examples
        # OHA no examples
        # OHB no examples
        # OHC no examples
        # OC2 no exmampes

        # Fallback oxygen
        ('O',   'O',   0),

        # Carbon SP
        ("CSP1", '[H][C]#*',  1), # e.g. in 2GT
        ("CSP",  '[C]#[C]',   (0,1)),
        ("CSP",  '[C]#*',     0),

        # Carbon SP2
        ('CR56', 'c12aaaac1aaa2',  (0,5)), # works on indole
        ('CR56', 'c12aaaan1aaa2',  0), # same pattern as (below) N56, but catching first 56 atom
        ('CR56', 'c12AAAAc1aaa2',  0), # 6-ring doesn't have to aromatic (8UG)
        ('CR66', 'c12aaaac1aaaa2', (0,5)),
        ('CR6',  'c12ccccc1OCO2',  (0,5)),   # mouse, fused atoms in 6-ring not non-arom 5-ring
        ('CR66', 'c12aaaac1AAAA2', (0,5)),   # one 6-ring aromatic, other not. Needed for XXX?
                                             # but makes a fail on 113.
        ('CR6',  'c12caccc1***2',  (0,5)),  # aromatic 6, (non-)aromatic 5, maybe this should be CR56?

        # note CR1  missing - can't find example
        #      CR1H missing - can't find example

        ('CR16', '[cr6;H1]',  0),
        ('CR6',  '[cr6;H0]',  0),
        ('CR15', '[cr5;H1]',  0),
#        ('CR5',  'C1(=O)[C,c][C,c]C=N1', 0), # carbonyl C in a (non-percieved?) 5 ring, 0CE (a)
        ('CR5',  '[cr5;H0]',  0),
        ('CR5',  '[CR5;H0]',  0),
        ('C1',   '[CX3;H1]',    0),  # double bond, single bond and one H
        ('C2',   '[CX3;H2]=*',  0),  # double bond, and 2 H
        ('C',    '[CX3;H0;^2]', 0),
        ('C',    '[CX3]=[OX1]', 0),  # carbonyl carbon
        ('C',    '[$([CX2](=C)=C)]',   0), # bonded to 3 things not hydrogen

        # Carbon SP3
        ('CT',   '[CX4H0]', 0),     # single bonded to 4 things not hydrogen
        ('CH3',  '[C;H3;^3]',   0), # bonded to something via single bond and 3 Hs
        ('CH2',  '[C;^3;H2]',   0), # standard aliphatic C.
        ('CH1',  '*[C;H1](*)*', 1), # bonded to H and 3 things

        # sp??? needs sorting
        ('CH2',  '[CH2]',   0), # bonded to 2 hydrogens

        # Carbon fallback
        ('C', '[C,c]', 0),

        # Hydrogen
        ('HCH1', '[H][CH1]',    0),
        ('HCH2', '[H][C;H2^3]', 0),
        ('HCH3', '[H][CH3]',    0),
        ('HNC1', '[H][N;H1;^2]~C(~N)~N', 0), # H of N of N=C ?
        ('HNC2', '[H][NX3;H2;^2]', 0), # H on a NC2 (NH1 and NH2 of ARG)
        ('HNC3', '[H][NX3;H3;^2]', 0), # guess - no examples
        ('HNT1', '[H][NX4;H1;^3]', 0),
        ('HNT1', '[H][NX3;H1;^3]', 0),
        ('HNT2', '[H][NX3;H2;^3]', 0), # H connected to type NT2
        ('HNT3', '[N^3;H3][H]', 1), # NH3+
        ('HNH2', '[H][NH2;^2]', 0), # NH2 is sp2
        ('HNH1', '[H][NX3;H1;^2]',    0),
        ('HCR6', '[H][cr6;H1]', 0),
        ('HCR5', '[H][cr5;H1]', 0), # connected to aromatic ring C with 1 H
        ('HNR5', '[H][nr5;H1]', 0), # connected to aromatic ring C with 1 H
        ('HNR5', '[H][Nr5;H1]', 0), # guess based on above, connected to aromatic N in a 5-ring
        ('HNR6', '[H][nr6;H1]', 0), # connected to aromatic 6-ring N with 1 H
        ('HNR6', '[H][NR6;H1]', 0), # guess based on above

        # HCR missing - no examples (and how is it different to HCR1?)
        ('HCR1', '[H]c',        0),
        ('HNH1', '[H][NH1]',    0),
        ('HOH1', '[H][OH1]',    0),
        ('HOH2', '[H][OH2][H]', (0,2)), # H of HOH - water
        ('H',    '[H]',         0),

        # Nitrogen, SP3

        ('NT1', '[NX4;H1;^3]',  0),
        ('NT1', '[NX3;H1;^3]',  0),
        ('NT2', '[NX3;H2;^3]',  0), # different to mon-lib!
        ('NT3', '[NX4;H3;^3]',  0),
        ('NT',  '[NX3;H0;^3]',  0),


        # NE-CZ in ARG should be deloc (guandino) - also NC1-C
        # single (as is in ARG.cif) is not found in ener_lib!

        # Nitrogen, SP2
        ('NR66', 'c12aaaan1aaaa2', 5), # (second) 66 atom is an N.
        ('NR56', 'c12aaaan1aaa2',  5), # (second) 56 atom is an N.
        ('NR55', 'c12aaan1aaa2',   4), # (second) 55 atom is an N.
        ('NC2',  '[NX3;H2^2]', 0),     # N of sp2 NH2 (as in ARG).
        ('NH2',  '[NX3^2][CX3^2]=[N^2;X3+]', (0,2)), # amidinium (charged)...
        ('NR15', '[nr5;X3;H1]',    0),
        ('NR5',  '[nr5;X3;H0]',    0),
        ('NR5',  '[NR;X3;H0;^2]',  0), # [NR5;X3;H0;^2] fails on 14C (also at daylight)
        ('NRD5', '[nr5;X2;H0]',    0), # guess from 071
        ('NRD5', 'C1(=O)[C,c][C,c]C=N1', 5), # N bonded to carbonyl C in a (non-percieved?) 5 ring, 0CE (a)
        ('NR16', '[nr6;H1]',    0),
        ('NRD6', 'a:[nr6;X2;H0]:a',  1), # aromatic N with no H, i.e. one double one single
        ('NR6',  '[nr6]',    0),
        ('NC1',  '[H][N;H1;^2]~C(~N)~N', 1),
        ('NC1',  '[NX3;H1;^2]C(~N)~N', 0), # N, as in NE in ARG
        ('NC1',  '[NX2;H1;^2]', 0),  # N of N=C ?
        ('NH1',  '[NX3;H1;^2]', 0),
        ('NH2',  '[NX3;H2;^2]', 0),  # sp2, e.g. ND2 of an ASP
        ('NT',   '*n1~[o]~[o]1', 1), # guess from 16X dioxaziridine (bleugh)
        # (NT needs checking?)
        # NC2 no examples
        # NC3 no examples
        # NPA no examples
        # NPB no examples


        # Nitrogen SP1
        ('NS',   '[N^1]', 0),
        # NS1 no examples


        # fall-back nitrogen
        ('N',    '[N,n]',      0),

        # Phosphorus
        ('P',    'P', 0),
        # Cl
        ('CL',   '[Cl]', 0),
        # F
        ('F',    '[F]',  0),
        # Br
        ('BR',    '[Br]',  0),

        # Sulfur
        ('SH1',  '[SX2H1]', 0),  # SG of CYS
        ('ST',   '[SX4]', 0),  # tetrahedral (2 single bonds, 2 double)
        ('S1',   '[S]=*', 0),
        ('S2',   '[SX2,sX2]', 0),
        ('S3',   '[SX3,sX3]', 0),
        ('S',    '[S,s]', 0),

        # Silicon
        ('SI1',  '[Si;X4]', 0), # tetragonal Si
        ('SI',   '[Si]',    0)  # Si any other

        ]

    full_list = smarts_list

    for item in smarts_by_element():
       full_list.append(item)

    # print('----------------------------refmac-----------------------------------------')
    for smarts_info in full_list:
        atom_type, smarts, match_atom_index = smarts_info
        pattern = Chem.MolFromSmarts(smarts)
        if mol.HasSubstructMatch(pattern):
            matches = mol.GetSubstructMatches(pattern)
            if False:
                print("SMARTS ", smarts)
                print("  ", atom_type, ": ", matches)
            for match in matches:
                set_atom_type(match, match_atom_index, mol, atom_type)
        else:
            # print "SMARTS ", smarts, " --- No hits  "
            pass

    # do we return success (everything has a type) or not?
    #
    for atom in mol.GetAtoms():
       try:
          atom_type = atom.GetProp('type_energy')
       except KeyError:
          is_aromatic = atom.GetIsAromatic()
          hybrid      = atom.GetHybridization()
          print("Error:: Missing type for atom \"", atom.GetProp('name'), "\" is_aromatic: ", is_aromatic, " hybridization: ", hybrid)
          return False
    # we got to the end, good
    return True

def set_amber_atom_types(mol):
    smarts_list = [

        ('C',   '[CX3]=[OX1]', 0), # sp2 C carbonyl group
        ('CA',  'c', 0), # aromatic C
        ('CB',  'c', 0), # aromatic C of fused 5,6 membered rings
        ('CC',  '[cr5;H1]', 0), # C in HIS
        ('CD',  'C=C-C=C', 1), # CD in C=CD-CD=C
        ('CK',  'n[cr5]n', 1), # C in the 5 membered ring of purines
        ('CM',  '[cr6;H1]n', 0), # C in pyrmidines, positions 5,6
        ('CM',  '[cr6;H1]c', 0), # C in pyrmidines, positions 5,6
        ('CN',  '[cr6]n[cr5]', 0), # aromatic C 5,6 membered rings. Is that enough?
        ('CQ',  'n[cr5]n', 1), # C of 5 membered ring in purines N-C-N
        ('CR',  'c[cr5]n', 1), # as above but in HIS
        ('CT',  '[CX4]', 0), # sp3 C aliphatic
        ('CV',  'c', 0), # arom 5 membered ring with 1 N and 1 H (HIS)
        ('CW',  'c', 0), # arom 5 membered ring with 1 N-H and 1 H (HIS)
        ('C*',  'c', 0), # arom 5 membered ring with 1 subst. (TRP)
        ('C*',  'c', 0),  # arom 5 membered ring with 1 subst. (TRP)
        ('CY',  'C#N', 0), # nitrile C
        ('CZ',  'A=C=A', 1), # sp C
        ('CZ',  'C#A', 1), # sp C

        ('O',  'O=C',  0), # carbonyl oxygen
        ('OH',  'COH', 1), # alcohol oxygen
        ('OS',  'C(=O)OA', 2), # ethyl or ester oxygen
        ('OW',  'HOY', 1), # water oxygen
        ('O2',  'O=CO', 1), # carboxyl oxygen
        ('O2',  'O=PO', 1), # phosphate

        ('SI',   '[Si]',    0)  # Si any other
        ]

    full_list = smarts_list
    for smarts_info in full_list:
        atom_type, smarts, match_atom_index = smarts_info
        pattern = Chem.MolFromSmarts(smarts)
        if mol.HasSubstructMatch(pattern):
            matches = mol.GetSubstructMatches(pattern)
            if False:
                print("Amber SMARTS ", smarts)
                print("  ", atom_type, ": ", matches)
            for match in matches:
                set_atom_type(match, match_atom_index, mol, atom_type)
        else:
            # print "SMARTS ", smarts, " --- No hits  "
            pass


def set_parmfrosst_atom_types(mol):

    electroneg      = '[N,n,O,o,F,S,s,Cl,Br]'
    ring_electroneg = '[N,n,O,o,s,S]' # are the non-arom needed?
    branched_electroneg = '(~' + electroneg + ')'
    g = '[!H&!He&!Ne&!Ar&!Xe&!Kr&!Rn]'
    g_and_ar = '[c,n,s]' # I believe
    # ar = in aromatic-5-ring or aromatic-6-ring
    # r3 = in a 3 membered ring: g~g~g~@1


    # print('debug H3a: {}'.format('[H][C^3;H1]'+branched_electroneg+branched_electroneg+electroneg))
    smarts_list = [

        # Hydrogen
        # initally from http://www.chem.cmu.edu/courses/09-560/docs/msi/ffbsim/B_AtomTypes.html
        #('H2', '[H][C^3;H1][O]', 0), # O is an example electroneg, more specific than H1
        ('HP',  '[H][C^3;X4][+]',    0), # H on a CT next to atom with formal positive charge
        ('H3a', '[H][C^3;H1]'+branched_electroneg+branched_electroneg+'~'+electroneg, 0),
        ('H2a', '[H][C^3;H1]'+branched_electroneg+electroneg, 0),
        ('H2a', '[H][C^3;H2]'+branched_electroneg+electroneg, 0),
        # ('H5',  '[H][c^2;H1]([n])[n]', 0),
        ('H5',  '[H][c^2;H1]'+branched_electroneg+'~'+electroneg, 0),
        ('H5',  '[H][C^2;H1]'+branched_electroneg+'~'+electroneg, 0),
        ('H1a', '[H][C^3;H1]'+electroneg, 0),
        ('H1b', '[H][C^3][N,n]',  0), # {aliph H; 1 elneg grp}
        ('H1e', '[H][C^3]' + electroneg,  0), # {aliph H; 1 elneg grp}
        ('H4a', '[H][C^2,c^2]'+electroneg, 0), # H4 is H-C&sp2~elneg
        ('H4b', '[H][C^2,c^2]~'+electroneg, 0),
        ('H4c', '[H][C^2,c^2;H1][n]', 0), # was '[H][c^2;H1][n][H]'
        ('H4d', '[H][c^2;H1][s]', 0), # not electroneg! parm_Frosst.pcp inconsistent!
        ('HA', '[H][C^2]',       0),
        ('HA', '[H][cr;^2]',     0), # aromatic H, HA also bonds to [C^2]
        ('H',  '[H][NX3;H1;^2]', 0), # both this  and the one below? Checkme
        ('H',  '[H][nX3;H1;^2]', 0),
        ('HC', '[H][CH3;^3]',    0),
        ('HC', '[H][c,CH1]',     0),
        ('HC', '[H][c,CH2]',     0),

        # Amber weirdness:
        # H-O&x2-C&sp3-O&x2-H        > HX * * * HX        ; {acetal hydroxyl must have vdW rad}
        # H-O&x2-C&sp3-N&sp3        > HX * * *        ; {acetal hydroxyl must have vdW rad}
        ('HXa', '[H]-[OX2]-[C^3]-[OX2]-[H]',  (0,4)), # these are by the book
        ('HXb', '[H]-[OX2]-[C^3]-[N^3]',      0),
        ('HXc', '[H]-[OX2]-[C^3](N)(N)',      0), # this is to match H14 in PF-113 -
                                                  # it's not the book acetal pattern
        ('HXe', '[H]-[OX2]-[C^3]([N^3])[c]', 0),  # ;H1 on N added to distinguish
                                                  # PF-345 and PF-261 (they look very close
                                                  # to me.
        ('HXf', '[H]-[OX2]-[C^3][N;X3]', 0), # this is acetal hydroxyl
        ('HXg', '[H]-[OX2]-[C^3][n;X3]', 0), # aromatic N

        ('HO', '[H][O;H1]',      0),
        ('HS', '[H][S]',         0),
        ('HW', '[H][O;H2]',      0),
        ('H2b', '[H][C^3]'+branched_electroneg+electroneg, 0), # not sure this works

        # fallback
        ('H', '[H]',    0),

        # ------------------------------- pathological specialities ------------------------------
        #
        # fused 5,6 unsat systems
        # the v and ws go together (the same SMART) for normal 5,6 ring and "his-like" sets:
        # normal 5,6 ring
        ('CBv', 'c12aaaac1ccn2',  (0,5)), # indole, hit the C[5,6] carbons
        ('CCv', 'c12aaaac1ccn2',  6),     # indole, hits a non-fused carbon
        ('CWv', 'c12aaaac1ccn2',  7),     # indole, hits a non-fused carbon

        # 5,6 { normal 5,6-ring } CB CB CC CW *, add in the branch so that it
        # doesn't hit sat carbon in 6-ring (this may not be the right thing to catch)
        ('CW-row-2', '[c;r5;R2]([C;X3;r6])[c;r5;R2]([C;X3])[c;r5][c;r5]'+electroneg, 3),
        ('CC-row-2', '[c;r5;R2]([C;X3;r6])[c;r5;R2]([C;X3])[c;r5][c;r5]'+electroneg, 2),
        ('CB-row-2', '[c;r5;R2]([C;X3;r6])[c;r5;R2]([C;X3])[c;r5][c;r5]'+electroneg, (0,1)),
        # catch C1 pf4 (hack?)
        ('CW-row-2b', '[c;r5;R2]1[c;r5;R2][c;r5][c;r5]'+electroneg+'1', 3),
        # PF_12
        ('CB-row-2c', '[c;r5]1:[c;r5]~[#6;r5]~[#6;r5][NH]1', (0,1)),
        # ('CC-row-2c', '[c;r5;r6;R2]1:[c;r5;r6;R2][#6;r5][#6;r5]'+electroneg+'1', 2),
        # ('CW-row-2c', '[c;r5;r6;R2]1:[c;r5;r6;R2][#6;r5][#6;r5]'+electroneg+'1', 3),

        # As it stands c12aaaac1aaa2 has ~50% wrong hits
        # ('CBw', 'c12aaaac1aaa2',  (0,5)), # hit the C[5,6] carbons # which way round do we go?
        # ('CCw', 'c12aaaac1caa2',  6),     # hits a non-fused carbon
        # ('CWw', 'c12aaaac1aca2',  7),     # hits a non-fused carbon

        # 5,6 ring with "HIS-like" 5-ring (does this hit anything? second ring is 5 (or 4???))
        ('CBx', '[cr5]12aaaa[c5]2-'+electroneg+'=[cr5]-'+ring_electroneg+'1', (0,5)),
        ('CRx', '[cr5]12aaaa[c5]2-'+electroneg+'=[cr5]-'+ring_electroneg+'1', 7),

        # 5-ring of indole:  { 5-ring unsat }
        ('CR-indole', '[c;r5;H1]1n[c;R2][c;R2]c1',  0),
        # ('NB-indole', 'c1nccc1',  1),

        # PF_5
        ('CCj', '[cr5]1(N)[cr5][ar5][cr5][cr5]1', 0),
        ('CWj', '[cr5]1(N)[cr5][ar5][cr5][cr5]1', 2), # account for branch atom

        # PF_3
        ('CWi', '[cr5]1[cr5][cr5][sr5][sr5]1', 0),
        ('CCi', '[cr5]1[cr5][cr5][sr5][sr5]1', 1),

        # real histidine-like
        # note: [nH] is not the matched by [n], this HIS Nitrogen atoms are different
        # and will match differently with different routes round the ring.
        #
        ('CCy',       '[cr5]1[cr5;H]'+ring_electroneg+'[cr5]'+ring_electroneg+'1', 1), # HIS
        ('CC-testc',  '[cr5]1[cr5;H][nr5][c][nH]1', 1), # HIS
        ('CC-testcc', '[cr5]1[cr5H][nr5;H][c][nr5]1', 1), # HIS, other way
        ('CWy', '[cr5]1[cr5]'+ring_electroneg+'[cr5]'+ring_electroneg+'1', 0), # HIS
        ('CWy', '[cr5]1[cr5]'+ring_electroneg+'[cr5]'+'[nH]1', 0), # HIS
        ('CRy', '[cr5]1[cr5]'+ring_electroneg+'[cr5]'+ring_electroneg+'1', 3), # HIS # success - all hits!
        #
        # maybe we should allow CWy to match CC and CCy to match CW - 2 way rounds an imidazole

        # * * * CR * ; { 5-ring unsat }
        ('CR-5ring-A', g+'1'+g+'N=C'+g+'1', 0), # what does this catch?
        ('CR-5ring-B', '[C;H0;X3]1=NC=CC1', 0), # H0 may not be needed.
        ('NB-5ring-B', '[C;H0;X3]1=NC=CC1', 1), # H0 may not be needed.

        # 5-ring unsat C&ar5=N&ar5-g&ar5~g&ar5~g&ar5-@1        -> CR NB * * * (above the comment)
        #
        # the C=N double bond above is tested for here by X2 on the N (atom index 1)
        ('CR-5ring-a',  '[C,c;^2;R1]1~[n;R1;X2]~[r5;^2;R1;!n]~[ar5;R1]~[C,c;^2;R1]1', 0),
        ('NB-5ring-a',  '[C,c;^2;R1]1~[n;R1;X2]~[r5;^2;R1;!n]~[ar5;R1]~[C,c;^2;R1]1', 1),
        # with a sp3 - bridging C atoms can be aromatic
        ('CR-5ring-b',  '[C,c;^2]1~[N]~[r5;^3]~[r5]~[C,c]1', 0),
        ('NB-5ring-b',  '[C,c;^2]1~[N]~[r5;^3]~[r5]~[C,c]1', 1),
        # for PF4 N4
        ('CR-5ring-c',  '[C,c;^2;R1]1~[n;R1;X2]~[a;r5;^2;R1]~[ar5;R2]~[a;^2;R2]1', 0),
        ('NB-5ring-c',  '[C,c;^2;R1]1~[n;R1;X2]~[a;r5;^2;R1]~[ar5;R2]~[a;^2;R2]1', 1),
        # for PF4 N6
        ('NB-5ring-d',  '[#6;R1]1~[#7;R1]~[#6;R2]~[#7;R2]~[#7;R1]1', 1),
        # for PF4 N5
        ('NB-5ring-d',  '[#6;R2]1~[#7;R1]~[#7;R1]~[#6;R1]~[#7;R2]1', 1),
        # for PF8 N3

        # Note to reader: compare N2 in PF11 and N2 in PF_15.  Extremely similar.  But different types
        #                         NB             NA
        # maybe because of the non-H substituents on atoms idx-3 and idx-4?
        # so add H1 to make it "his-like" - testing.

        # ('NB-5ring-e',  '[c;R1]1~[n;R1]~[a;R1]~[a;R1]~[a;R1]1', 1), # idx-2 atom not a (too general)
        ('NB-5ring-e2', '[c;R1]1~[n;R1]~[n;R1]~[c;R1;H1]~[c;R1;H1]1', 1), # idx-2 atom not a
        ('NB-5ring-e3', '[c;R1]1~[n;R1]~[c;R1]~[n;R1]~[s;R1]1', 1), #
        # for PF8 N4
        ('NB-5ring-f',  '[c;R1]1[n;R1][s;R1][a;R1][a;R1]1', 1), # atom idx-2 needs to be "s" not "a"
        # for PF9 N4
        ('NB-5ring-g',  '[c;R1]1~[n;R1]~[n;R2]~[a;R2]~[c;R1]1', 1),
        # note to self - there may be more CR NBs
        ('CR-5ring-g',  '[c;R1]1~[n;R1]~[n;R2]~[c;R2]~[c;R1]1', 0),
        ('CR-5ring-h',  '[c;R1]1~[n;R1]~[n;R1]~[c;R1]~[c;R1]1', 0), # PF11
        ('NB-5ring-i',  '[c;R1]1~[n;R1;H0]~[n;R2;X3]([c;r6])~[c;R2]~[n;R1;H0]1', 1), # N3 in PF14,  but not N2 in PF_4
        ('NB-5ring-j',  '[c;R1]1~[n;R1]~[n;R1]~[c;R1]~[c;R1]1', 2), # N1 in PF15?

        # 5-ring unsat C&ar5=C&ar5-C&ar5=C&ar5-g&ar5-@1 -> CW CC CC CW
        ('CW-5ring-b',  '[cr5]1[cr5][cr5][cr5][ar5]1', (0,3)),
        ('CC-5ring-b',  '[cr5]1[cr5][cr5][cr5][ar5]1', (1,2)),

        # 5-ring unsat C&ar5=C&ar5-C&ar5=C&ar5-g&ar5-@1        > CW CC CC CW
        ('CWc',  '[cr5;R1]1[cr5][cr5][cr5][nr5]1', (0,3)),
        ('CCc',  '[cr5;R1]1[cr5][cr5][cr5][nr5]1', (1,2)),
        ('CWg',  '[cr5;R1]1[cr5][cr5][cr5][ar5]1', (0,3)),
        ('CCg',  '[cr5;R1]1[cr5][cr5][cr5][ar5]1', (1,2)),

        # 5-ring unsat CW CC * * *, Note CR NB is higher than this
        ('CW-5ring-c',  '[cr5]1[cr5][cr5][cr5][nr5]1', 0),
        ('CC-5ring-c',  '[cr5]1[cr5][cr5][cr5][nr5]1', 1),
        ('CW-5ring-d',  '[c;r5]1[c;r5][c;r5][n;r5][n;r5]1', 0),
        ('CW-5ring-d',  '[c;r5]1[c;r5][c;r5][n;r5][n;r5]1', 2), # symmetry of above
        ('CC-5ring-d',  '[c;r5]1[c;r5][c;r5][n;r5][n;r5]1', 1),

        # 5-ring unsat, NB has no H, NA has H, i.e. HIS C=N-C vs C-N(H)-C
        ('NBz',  '[nr5;R1]1[cr5;R1][cr5R1][nr5;R1;X2][ar5R1]1', 3),
        ('NAz',  '[nr5;R1]1[cr5;R1][cr5R1][nr5;R1;X2][ar5R1]1', 0), # does this find anything?
        ('CRz',  '[nr5;R1]1[cr5;R1][cr5R1][nr5;R1;X2][ar5R1]1', (1,2)),

        # ('CR-5-ring-j', '[c]1[n;H][n][c][c]1', 0), # catches C2 in PF_15 # too late
        # ('CW-5-ring-j', '[c]1[n;H][n][c][c]1', 3), # catches C4 in PF_15

        # n1cnnc1
        # ('NBzz',  '[nr5;R1]1[cr5;R1][nr5R1][nr5R1][ar5R1]1', (2,3)),
        # ('CRzz',  '[nr5;R1]1[cr5;R1][nr5R1][nr5R1][ar5R1]1', 1),
        # we can be more specific and catch the other carbon:
        # we may need to generalize again (just comment above?)
        ('NBzz',  '[nr5;R1]1[cr5;R1][nr5R1][nr5R1][cr5R1]1', (2,3)),
        ('CRzz',  '[nr5;R1]1[cr5;R1][nr5R1][nr5R1][cr5R1]1', (1,4)),

        # non-aromatic (a ring carbon atom has 2 hydrogens)

        # 5-ring unsat, NB has no H, NA has H, i.e. HIS C=N-C vs C-N(H)-C
        ('NBna',  '[N;r5;R1]1=[C;r5;R1][C;r5;R1][N;r5;R1;X2]=[A;r5;R1]1', (0,3)),
        ('CRna',  '[N;r5;R1]1=[C;r5;R1][C;r5;R1][N;r5;R1;X2]=[C;r5;R1]1', 4), # PF20, non-arom though

        # ------------------------------ back to sanity ---------------------------

        # Carbon

        # carbon order: CP {rings/fused-rings} C CR CB C* CA CM C2 CJ CT
        #

        ('CPa', '[cr6]-[cr6]', 0), # biphenyl bridge
        ('CPb', '[cr5]-[cr5]', 0), # biphenyl bridge
        ('CPe', 'c1c(cccccc)aacc1', 1), # biphenyl bridge, hits C7 in PF23

        # ('CCb',  '[R]1:[c]:[R]:[R]:[R]:1', 1), # hits things that should be CR or CW
        ('CCc',   '[cr5][c;H0;X3;r5][cr5][nr5][cr5]', 1),

        # ('CCd', '[cr5;X3;^2]', 0), # 5 ring unsaturated

        ('CRa',   '[cr5]1[nr5;H0][ar5][ar5][ar5]1', 0), # this is useful
        ('CRb',   'n[cr5;R2]n', 1),
        ('CRc',   'n[cr5;R1;H0]n', 1),
        ('CRd',   'n1[cH1]ncc1', 1), # NE2 HIS

        # ('Ca',  '[c]:[cX3]=[N]', 1), # was '[c]~[cX3]=[N]' for C10 in PF7, but hits CRs
        ('Ca',  '[C]-[CX3]=[N]', 1),
        ('Cb',  '[CX3]=[O]',     0),
        ('Cc',  '[cX3]=[O]',     0),
        ('Cd',  '[H][CR0]=[N]',  1),
        ('Ce',  '[H][CR0]=[N]',  1), # also conyl
        ('Cf',  '[!C][CR0]=[N]', 1), # also conyl

        # CB is fused aromatic bridge-head
        ('CBa', 'c12aaaac1aaa2', 0), # doesn't hit anything
        ('CBb', 'c12aaaac1aaaa2', (0,5)),
        ('CBc', 'c12aaaac1aan2',  (0,5)), # indole

        # atom CG of a TYR has type CA
        #
        ('CAa', '[a][cr6;H0;R1]([C,c,N,n])[a]', 1),
        ('CAb', 'c', 0),  # CA on fusion atoms in PF6
        ('CAc', '[cr6;X3]', 0),
        ('CAd', '[cr6;H1]', 0),
        ('CAe', '[cr6]',    0), # Does this work? (PF6)
        ('CAf', '[Cr6]',    0), #  [CR6] means in 6 rings :-)
        ('CAg', 'c1ccccc1',  (0,1,2,3,4,5)),
        ('CAh', 'c1ccccn1',  (0,1,2,3,4)),
        ('CAg', '[C^2]1~[C^2]~[C^2]~[C^2]~[C^2]~[C^2]1',  (0,1,2,3,4,5)),
        ('CAh', '[C^2]1~[C^2]~[C^2]~[C^2]~[C^2]~N1',  (0,1,2,3,4)),

        # SMARTS put an atom in a 5 ring before a 6 ring. but CA is 6-ring
        ('C*', '[cr5;^2;X3]', 0),
        ('C*', '[C;X3;^2]1C(=O)[NH]~cc1', 0), # C8 in PF_12 is not aromatic (by rdkit) but is a PF C*. Grr!
                                              # So provide the specified ring.

        ('CMa', 'n1cnccc1', (3,4,5)), # positions 5 or 6 in pyrimidine # or 4 presumably!
        ('CMb', '[C^2]=[C^2]', (0,1)),  # by validation # catches C7 in PF_0 {non-aromatic C=x}
        ('CMb', '[C;^2;R0]=[C^2]', 0),  # by validation
        ('CMc', '[C;^2;R0]=A', 0),   # by validation
        ('CMd', '[C]=a', 0),   # by validation

        ('C2', '[CX2]#[CX2]', (0,1)),   # by validation
        ('C2', '[CX2]#A', 0),   # by validation
        ('C2', 'A#[CX2]', 1),   # by validation [1]

        ('CWd', '[cr5;X3;^2][c]=O', 0),
        ('CWa', 'n[cr5;R2][cr5;R2]c', 1), # C4 in PF4

        # cyclopropyl and epoxide
        ('CJa',   '[CX4]1[CX4][CX4]1',  (0,1,2)),
        ('CJb',   '[CX4]1[CX4]A1', (0,1)), # remove unindex symmetry

        ('CT', '[CX4]', 0), # bonded to 4 things
        # Carbon fallback
        ('C-fallback',  '[C,c]', 0), # sp hybrizided. Hmmm.

        # [1] needed because above [CX2]#A doesn't find the bond both ways.  I suppose
        #     that's the way SMARTS work.


        # Oxygen
        # first is the carboxylate special:
        ('O2-carboxylate', '[C,c][CX3](=O)[O-]', (2,3)),
        # and presumably this too, but not come across it yet)
        # ('O2-carboxylic-acid', '[C,c][CX3](=O)[OH1]', (2,3)),
        # what's the point of using R0 and R1 and R2!?
        ('OSa',  "[OX2;H0;R0]", 0), # ester
        ('OSb',  "[OX2;H0;R1]", 0), # ester
        ('OSc',  "[OX2;H0;R2]", 0), # ester PF-85 - in two rings!
        ('OSd',  "[oX2;H0;r5]", 0), # aromatic, should I add [n]?
        ('OSe',  "[O;-][CX4]", 0),  # alkoxide ion
        ('OSf',  "[cX3][o][cX3](=O)c", 1), # aromatic in ring with ketone oxygen

        # this is too liberal - it matches O1 in PF 319 (10 other good matches though)
        # ('OSd6', "[oX2;H0;r6]", 0), # aromatic, should I add [n]?
        ('OSd6a', "[oX2;H0;r6]1[cR2][cR2]c(=O)cc1", 0), # a couple
        ('OSd6b', "[oX2;H0;r6]1[c](=O)[n]ccc1", 0), # PF-115
        ('OSd6c', "[oX2;H0;r6]1ccc(=O)cc1", 0), # PF-333
        ('OSd6d', "[oX2;H0;r6]1ccc(=O)cc1", 0), #

        # this matches O1 in 319 - bad!
        # ('OSf',  "[cX3]1[o][cX3;R2][cX3R2]cc1", 1), # aromatic in ring with ketone oxygen
        ('OSg',  "[cX3]1[o][cX3;R2][cX3R2]c(=O)c1", 1), # aromatic in ring with ketone oxygen
        ('OSh',  "[cX3]1[o][cX3][nX3]c(=O)c1", 1), # PF-76 ring is aromatic
        ('OSi',  "[cX3]1[o][cX3](=N)ccc1", 1), # PF7


        # should be 'O'? c.f. 319 vs 333
        #('OSg',  "[cX3]1[o][cX3;R2][cX3R2]cc1", 1), # aromatic in ring with ketone oxygen

        ('OW',  "[OH2]", 0), # water
        ('OH',  "[OH1]", 0), # alcohol

        ('O2',  "[O;-]-*", 0), # oxide ion - deloc
        # O2 should match sulphate (and phosphates, I think) but not
        # O=S-{non-oxygen}
        ('O2',  'OP(~O)~O',   0), # O on a phosphate
        ('O2',  'OS(~O)~O',   0), # guess based on above
        ('O2',  'O=P',   0),      # O on a phosphate
        ('O2',  'O=S=O',   (0,2)), # guess based on above

        # horrible hacking c.f. PF2:  O1=S1 -> O
        #                       PF56  O1-S1 -> Ou
        ('O',   'O=[C]-*',   0), # carbonyl oxygen
        ('O',   'O=[c]',     0), # weird ring iiin PF-3
        ('O',   'O=[S;X4]C',   0),
        ('O',   'O=[S;X3]([N])C',   0), # weird
        # this looks very like one of the OSs to me.
        ('O',   'c12occcc1cccc2',   1), # hit PF-319, seems pretty perverse to do this though.
                                        # why isn't it another "ester"? We have
                                        # to have ultra-specific OS types so that
                                        # we don't hit this - grump.
        # fallback
        ('Ou',  'O',   0),
        ('Ou',  'o',   0),


        # Nitrogen
        #
        # Order: NJ NL N3 NC N2 NB N N2 N NA N3 N2 N3 ND N2 N N* N3 Nu
        #
        # N(-H)-C=O,S                > N * * *        ; {explicitly reset amide N to type N}

        # N=N-C&x4-@1 > NJ NJ *        ; {aziridine}
        ('NJ',    'N=N[CX4]',      (0,1)),

        # N&x1#g > NL *                ; {univalent triple-bonded N}
        ('NL',   '[NX1]#A',   0), # triple bond N

        # N&x4 > N3                ; {+ve tetrahedral N}
        ('N3b',   '[N^3;X3]',    0),
        ('N3c',   '[NX4;+]',     0),

        # g=N&ar6-g                > * NC *        ; {pyridine N}
        ('NC',   '[nr6;X2;H0]',  0), # pyridine

        # N&x3(-H)-Acy2-C         > N2 * C        ; {protonated imine}
        ('N2b',   '[NX3;H](=C)C',   0),   # fixme - don-t forget the C here too.

        # N&x3-C!ar=N&x3                > N2 C N2        ; {amidine}
        ('N2ax',   'A[C;^2](=N)N',   (2,3)),
        ('Cax',    'A[C;^2](=N)N',   1),  # NOTE! C here

        # Nitro > N2 ; {nitro group - guess jan 1999 CIBayly}
        ('N2f', 'C[N^2](~[O;X1])~[OX1]', 1),

        # g&ar-N&ar(-H)-g > * NA * * ; {e.g. HIS C-NH-C}
        # I don't understandy why this above pattern doesn't match N3 in PF_15.  Instead that
        # falls down to N (an amino N). Grump.
        # ('NAe', 'A[nH]a', 1), # too liberal, catches N3 in PF_15
        ('NAf', '[a][n;H;r5][a]', 1), # don't match nH in 6 rings - baah.

        # Nsulfon > N3 ; {sulfonamide-type N is pyramidal. Ordered after aromatic amine}
        # N&x3-S,P(~O&x1)~O&x1
        ('N3a', '[nX3]-[S]([OX1])~[OX1]', 0),
        ('N3b', '[nX3]-[S](=[OX1])~[OX1]', 0),
        # N&x2&fcm1-S,P(~O&x1)~O&x1
        ('N3c', '[nX2;+]-[S,P]([OX1])~[OX1]', 0),
        ('N3d', '[nX2;+]-[S,P](=[OX1])~[OX1]', 0),

        # g&ar-Acy1-N(-H,g)-g&ar > * N2 * * ; {more delocalized aniline-type N oct2004 CIBayly}
        # ('N2d', '[a][n;H]a', 1), # was ('N2d', '[a][N;H]a', 1), # not this - catches N3 in PF_15 (should be N)
        # ('N2e', '[a][n](-[a,A])a', 1), # not this - catches N3 in PF_15 (should be N)

        # g&ar-Acy1-N(-!H)-!H        > * N3 * *      ; {disubst aniline-type N apr2005 CIBayly}
        ('N3d', 'a[N;H0]([!H])[!H]', 1),

        # g&sp2-N!ar-H        > * N  * ; {sp2 amino N}
        ('N', '[A;^2][N][H]', 1),
        ('N', '[c][nH][c](=O)', 1), # N3 in PF_15

        # g&ar-N!ar > * N2 ; {sp2 amino N}
        ('N2g', g_and_ar + '-[N]', 1),

        # N&ar > Nstar                ; {sp2 N}
        ('N*', 'n', 0),

        # N&sp3        > N3                ; {regular sp3 N, not charged}
        ('N3', '[N;0;^3]', 0),

        # fall-back nitrogen
        ('Nu',    '[N,n]',      0),

        ('Sua', 'N[S](=O)(O)=N', 1), # Hmm! PF56

        # many flavours of hypervalent sulfur
        #
        ('SOa', '[S](=O)[N]', 0),  # hypervalent sulfur
        ('SOb', '[C^3][S](=O)(=O)[C^2,c^2]', 1),  # hypervalent sulfur, matches PF-22, 22 is aromic C in position 4, 75 is not (although it looks close)
        ('SOc', '[c][S](=O)(=O)[c]', 1),  # hypervalent sulfur
        ('SOd', '[c]-[S;X4](=O)(C)=[N]', 1),   # hypervalent sulfur
        ('SOe', '[n]-[S;X4](=O)(=O)[C]', 1),   # hypervalent sulfur
        ('SOh', '[c]-[S;X4](=O)(=O)[c]', 1),   # should hit 699, 567
        ('SOi', '[c]-[S;X4](=O)(=O)[O]', 1),   # should hit 521
        ('SOj', '[C;^3][S;X4](=O)(=O)[O]', 1), # hits 416

        ('Sa',  '[S,s][S,s]', (0,1)), # sulfide
        ('Sb',  '[s]', 0), # "sulfide" - hmm. PF5
        ('Sc',  '[c,C]S[c]', 1), # PF15
        ('Sd',  '[c,C]SN', 1), # PF85

        # let's try something less specific than that
        ('Se',  '[S]1[CX3][AX2]=[AX3][C]1', 0), # does this hit anything?
        ('Sf',  '[S]1[CX3]=[A][A][C]1', 0), # does this hit anything?
        ('Sg',  '[C^3][SX2][n]', 1), # should hit 609
        ('Sh',  '[C^3][SX2][c]', 1), # should hit 598
        ('Si',  '[C^3;R1]1[SX2;R1][CR1]AAA[C^2]1', 1), # should hit 496

        # sulfur
        ('Sub', 'S', 0),
        ('Suc', 's', 0), # yikes

        # P
        ('P',    'P', 0),
        # Cl
        ('CL',   '[Cl]', 0),
        # F
        ('F',    '[F]',  0),
        # Br
        ('BR',   '[Br]',  0),

        ]

    for smarts_info in smarts_list:
        atom_type, smarts, match_atom_index = smarts_info
        # print("SMARTS: {}".format(smarts))
        pattern = Chem.MolFromSmarts(smarts)
        if mol.HasSubstructMatch(pattern):
            matches = mol.GetSubstructMatches(pattern)
            if False:
                print("SMARTS ", smarts)
                print("  ", atom_type, ": ", matches)
            for match in matches:
                # print('set_atom_type', match, match_atom_index, mol, pattern, atom_type)
                set_atom_type(match, match_atom_index, mol, atom_type, types_type='Parm@Frosst')
        else:
            # print "SMARTS ", smarts, " --- No hits  "
            pass

    # do we return success (everything has a type) or not?
    #
    for atom in mol.GetAtoms():
       try:
          atom_type = atom.GetProp('pf_atom_type')
       except KeyError:
          is_aromatic = atom.GetIsAromatic()
          hybrid      = atom.GetHybridization()
          print("Error:: Missing type for atom ", atom.GetProp('name'), " is_aromatic: ", is_aromatic, " hybridization: ", hybrid)
          return False
    # we got to the end, good
    return True



#!/usr/bin/env python
"""Calculate the probabilities of various models for a given distance measure using linear interpolation or extrapolation"""

# Copyright 2010, 2011 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from math import floor


import sys
#from pprint import pprint

#from linearInterp import LinearInterp

DELIM = ","

class SmoothProb:
    """Calculate the probabilities of various models for a given distance measure using linear interpolation or extrapolation"""
    
    def __init__(self, fileName):
        """Creates and initialize a smoothProb object
        
        ARGUMENTS:
            inputFile - a file containing smoothed training data
        RETURNS:
            an initialized SmoothProb object
        """
        
        self.__xStart   = None #the starting value for the x-axis
        self.__xStep    = None #the size of the increment along the x-axis
        self.__fileName = None #the input filename
        self.__yData    = {}   #the y values read from fileName
        self.__yLen     = None #the number of Y values 
        self.__labels   = None #the names of the columns
        self.__fileName = fileName
        
        input = open(fileName, 'r')
        
        #read in the header line, which contains the labels for our probabilities
        labels = input.readline().split(DELIM)
        labels.pop(0) #we don't care about the header for the first column, since that's just distance
        labels = [x.strip() for x in labels] #get rid of any leading and trailing spaces (at the very least, get rid of the \n at the end of the line)
        self.__labels = labels
        
        #read in the first line, which determines xStart
        curdata = input.readline().split(DELIM)
        self.__xStart = float(curdata.pop(0))
        #create arrays for all y values
        for curindex, curvalue in enumerate(curdata):
            self.__yData[labels[curindex]] = [float(curvalue)]
        
        #read in the second line, which determines xStep
        curdata = input.readline().split(DELIM)
        self.__xStep = float(curdata.pop(0)) - float(self.__xStart)
        for curindex, curvalue in enumerate(curdata):
            self.__yData[labels[curindex]].append(float(curvalue))
        
        ep = self.__xStep / 10000 #since the step size is a float, we can't test that each step size is exactly
                                  #right without running into rounding issues.
                                  #Instead, we make sure that the step size is within .01% of the correct size
                                  #using this epsilon value
        
        #go through the rest of the input file
        for curline in input:
            #skip blank lines
            if len(curline.rstrip()) == 0:
                continue
            curdata = curline.split(DELIM)
            
            #make sure that step size is correct
            xVal = self.__xStart + self.__xStep * len(self.__yData)
            curdist = curdata.pop(0)
            if (xVal - ep <= curdist) and (curdist <= xVal + ep):
                raise "Incorrect smoothProb step size in file " + self.__fileName
            
            for curindex, curvalue in enumerate(curdata):
                self.__yData[labels[curindex]].append(float(curvalue))
        
        self.__yLen = len(self.__yData[labels[0]])
        
        input.close()
        
    
    def calcProb(self, xVal, label = None):
        """calculate the probabilities for a given distance
        
        ARGUMENTS:
            dist   - the distance
        OPTIONAL ARGUMENTS:
            label  - what model to calculate the probability for
                     must match a label in the first line of the input file
        RETURNS:
            if label is given, the probability that the model label is correct
            if label is not given, a dictionary containing probabilities for all models
        """
        
        xPos = int(floor( (xVal-self.__xStart) / self.__xStep ))
        
        #go through each label and determine a raw probability
        yDict = {}
        if xPos < 0:
            #if we're extrapolating down
            numSteps = (xVal-self.__xStart) / self.__xStep
            for curlabel in self.__labels:
                yStep = self.__yData[curlabel][0] - self.__yData[curlabel][1]
                yDict[curlabel] = self.__yData[curlabel][0] - numSteps * yStep
        elif xPos > self.__yLen-1:
            #if we're extrapolating up
            numSteps = (xVal-self.__xStart) / self.__xStep - (self.__yLen - 1)
            for curlabel in self.__labels:
                yStep = self.__yData[curlabel][-1] - self.__yData[curlabel][-2]
                yDict[curlabel] = self.__yData[curlabel][-1] + numSteps * yStep
        else:
            #if we're interpolating
            remainder = xVal - self.__xStart - self.__xStep * xPos
            for curlabel in self.__labels:
                yDict[curlabel] = self.__yData[curlabel][xPos] * ( 1 - remainder/self.__xStep) + self.__yData[curlabel][xPos+1] * (remainder/self.__xStep)
        
        #scale the raw probabilities so they add up to 1
        totalProb = sum(yDict.values())
        
        try:
            yDict = dict([(curlabel, curval/totalProb) for (curlabel, curval) in yDict.iteritems()])
        except ZeroDivisionError:
            #if totalProb = 0, then we're trying to make predictions with *very* bad data
            raise SmoothProbError("Underflow error: all probabilities equal to zero.")
        
        #return either the desired probability or the hash containing all probabilities
        if label is None:
            return yDict
        else:
            return yDict[label]


class SmoothProbError(Exception):
    #this error gets raised when all probabilities are 0
    def __init__(self, description):
        self.value = description
    def __str__(self):
        return repr(self.value)




import os
import coot
import math

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt


def cairoplot_pathology_plots(mtz, fp, sigfp):
    import cairo
    import cairoplot
    try:
        print("getting data for", mtz)
        data = coot.pathology_data(mtz, fp, sigfp)
        # print len(data), data
        prefix,tail = os.path.splitext(mtz)
        for i in range(4):
            png_file_name = prefix + "-" + str(i) + ".png"
            print( "making plot", png_file_name)
            yt = 'I'
            xt = 'Resolution'
            if i == 1:
                yt = 'I/SIGI'
            if i == 2:
                yt = 'SIGI'
            if i == 3:
                yt = 'I/SIGI'
                xt = "I"
            cairoplot.scatter_plot(png_file_name,
                                   data = data[i+1],
                                   # series_colors = ['black'],
                                   series_colors = [[0.2, 0.2, 0.2]],
                                   width = 600, height = 600, border = 20,
                                   discrete = True,
                                   x_title = xt,
                                   y_title = yt,
                                   dots = True,
                                   axis = True)

    except TypeError as e:
                print( "caught TypeError:", e)

# these can be I, SIGI also
#
def pathology_plots(mtz, fp, sigfp):

# these can be I, SIGI also
#
def plots(mtz, fp, sigfp):

    # meta data
    intensity_data = False
    if fp[0] == 'I':
        intensity_data = True

    data = pathology.PathologyData(mtz, fp, sigfp)
    # try:
    if True:
        prefix,tail = os.path.splitext(mtz)
        fp_lab = "FP"
        sigfp_lab = "SIGFP"
        if (intensity_data):
            fp_lab = "I"
            sigfp_lab = "SIGI"

        labels = [("Resolution", fp_lab),
                  ('Resolution', fp_lab + '/' + sigfp_lab),
                  (fp_lab, sigfp_lab),
                  (fp_lab, fp_lab + '/' + sigfp_lab)]
        z = zip(range(4), labels)
        if True:
            ii = 0 # label indexing hack
            l_0 = labels[ii][0].replace('/', '_')
            l_1 = labels[ii][1].replace('/', '_')
            png_file_name = prefix + "-" + l_0 + "-vs-" + l_1 + ".png"

            n_x_labels = 5
            """
            if l_0 == "Resolution":
                for j in range(n_x_labels):
                    f = data.invresolsq_max() * j / n_x_labels
                    if f == 0:
                        x_label = "Inf"
                    else:
                        x_label = str(round(math.sqrt(1/f), 3));
                    # print j, x_label
                    x_labels[j] = x_label
            else:
                x_labels = None
            """

            fig1 = plt.figure()
            xdata = [x[0] for x in data.fp_vs_reso()]
            ydata = [y[1] for y in data.fp_vs_reso()]

            ymin = 0
            if (intensity_data):
                ymin = min(ydata)

            xmax = max(xdata)
            ymax = max(ydata)
            plt.xlim(0, xmax)
            plt.ylim(ymin, ymax)

            s = 8
            # If this is F data the the y axis min should be 0.
            # The x-axis min should be 0.
            print("making plot", png_file_name)
            l = plt.scatter(xdata, ydata, s, alpha=0.2)
            plt.savefig(png_file_name)


    # except TypeError as e:
    #            print "caught TypeError:", e



#    begin.py
#    Copyright (C) 2008  Bernhard Lohkamp, The University of York
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

print("===============================================================")
print("==================== Testing ==================================")
print("===============================================================")

import coot
import sys
import unittest, os
import inspect

global have_test_skip
global skipped_tests

have_test_skip = False
skipped_tests  = []

if ('skip' in dir(unittest.TestCase)):
    have_test_skip = True
else:
    print("WARNING:: unittest skip not avaliable!!!!!!")

home = os.getenv('HOME')
if ((not home) and (os.name == 'nt')):
    home = os.getenv('COOT_HOME')
if (not home):
    # badness we dont have a home dir
    print("ERROR:: Cannot find a HOME directory")


def get_unittest_data_dir():
    d = os.getenv("COOT_TEST_DATA_DIR")
    if (not d):
        d = os.path.normpath(os.path.join(home, "data", "greg-data"))
    return d

global unittest_data_dir
unittest_data_dir = get_unittest_data_dir()   # too lazy to convert all
                                              # occurences of unittest_data_dir

################
# some functions
################

def unittest_pdb(file_name):
    ret = read_pdb(os.path.join(unittest_data_dir, file_name))
    return ret

# use functions rather than variables which would have to be globals
def rnase_pdb():
    return os.path.join(unittest_data_dir, "tutorial-modern.pdb")

def rnase_mtz():
    return os.path.join(unittest_data_dir, "rnasa-1.8-all_refmac1.mtz")

def rotate_n_frames(n):
    rotate_speed = 1
    return int(rotate_speed * n)

# whatever this shall do
#def chains_in_order_qm(ls):

def bond_length(pos_1, pos_2):
    def square(x):
        return x*x
    def sub(x, y):
        return x-y

    import math
    ret = math.sqrt(sum(map(square, list(map(sub, pos_1, pos_2)))))
    return ret

pos_diff = bond_length

def bond_length_from_atoms(atom_1, atom_2):
    from types import ListType
    if (type(atom_1) is not ListType):
        print("   WARNING:: bond_length_from_atoms: atom_1 not a list:", atom_1)
        return False
    elif (type(atom_2) is not ListType):
        print("   WARNING:: bond_length_from_atoms: atom_2 not a list:", atom_2)
        return False
    else:
        return bond_length(atom_1[2],
                           atom_2[2])

def bond_length_within_tolerance_qm(atom_1, atom_2, ideal_length, tolerance):

    # BL says:: this should (as in greg) probably throw an error, but we dont have the
    # test object here, so cannot. Buh. Should pass self?!
    if (not atom_1):
        return False
    if (not atom_2):
        return False
    b = bond_length_from_atoms(atom_1, atom_2)
    return abs(b - ideal_length) < tolerance

def shelx_waters_all_good_occ_qm(test, imol_insulin_res):

    chain_id = water_chain(imol_insulin_res)
    n_residues = chain_n_residues(chain_id, imol_insulin_res)
    serial_number = n_residues - 1
    res_name = resname_from_serial_number(imol_insulin_res, chain_id, serial_number)
    res_no   = seqnum_from_serial_number (imol_insulin_res, chain_id, serial_number)
    ins_code = insertion_code_from_serial_number(imol_insulin_res, chain_id, serial_number)

    if (res_name == "HOH"):
        atom_list = residue_info(imol_insulin_res, chain_id, res_no, ins_code)
        for atom in atom_list:
            occ = atom[1][0]
            test.assertAlmostEqual(occ, 11.0, 1, "  bad occupancy in SHELXL molecule %s" %atom)

#  return restraints without the given bond restraints or
# False if no restraints given
def strip_bond_from_restraints(atom_pair, restraints):

    import copy
    restr_dict = copy.deepcopy(restraints)
    if (restr_dict == []):
        return []
    else:
        if not restr_dict:
            return False
        else:
            # make a copy of restraints, so that we dont overwrite
            # the original dictionary
            bonds = restr_dict["_chem_comp_bond"]
            if bonds:
                for i, bond in enumerate(bonds):
                    atom1 = bond[0]
                    atom2 = bond[1]
                    if (atom1 in atom_pair and atom2 in atom_pair):
                        del restr_dict["_chem_comp_bond"][i]
                        return restr_dict
                return restr_dict

# are the attributes of atom-1 the same as atom-2? (given we test for
# failUnless/IfAlmostEqual)
#
def atoms_match_qm(atom_1, atom_2):
    from types import ListType, StringType, FloatType, IntType, BooleanType
    if ((not atom_1) or (not atom_2)):
        return False     # no matching residues
    if (len(atom_1) != len(atom_2)):
        return False     # comparing differnt list sizes -> not equal anyway
    for i in range(len(atom_1)):
        if (type(atom_1[i]) is ListType):
            if (atoms_match_qm(atom_1[i], atom_2[i])):
                pass
            else:
                return False
        elif (type(atom_1[i]) is StringType):
            if (atom_1[i] == atom_2[i]):
                pass
            else:
                return False
        elif (type(atom_1[i]) is FloatType):
            if (abs(atom_1[i] - atom_2[i]) < 0.01):
                pass
            else:
                return False
        elif (type(atom_1[i]) is IntType):
            if (atom_1[i] == atom_2[i]):
                pass
            else:
                return False
        elif (type(atom_1[i]) is BooleanType):
            if (atom_1[i] == atom_2[i]):
                pass
            else:
                return False
    return True


# transform
# Just a (eg 3x3) (no vector) matrix
# The matrix does not have to be symmetric.
#
def transpose_mat(mat, defval=None):
    if not mat:
        return []
    return list(map(lambda *row: [elem or defval for elem in row], *mat))

# What is the distance atom-1 to atom-2? 
# return False on not able to calculate
def atom_distance(atom_1, atom_2):
    def square(x):
        return x*x
    def sub(x, y):
        return x-y
    import math
    ret = math.sqrt(sum(map(square, list(map(sub, atom_1[2], atom_2[2])))))
    return ret

# a function from 04_cootaneering:
#
# note that things could go wrong if there is a mising EOL (not tested)
# not sure about this in python
#
def file2string(rnase_pir):
    fin = open(rnase_pir, 'r')
    s = fin.read()
    fin.close()
    return s

def atoms_have_correct_seg_id_qm(atoms, seg_id):

    for atom in atoms:
        atom_seg_id = atom[1][3]
        if (not atom_seg_id == seg_id):
            return False
    return True

# return the residue name given a 4 element spec
#
def residue_name_from_spec(imol, spec):
    
    if len(spec) == 4:
        return residue_name(imol, *spec[1:4])
    return False

# return residue specs of all residues that are type residue-type
#
def get_residues_in_molecule_of_type(imol, residue_type):

    return [x for x in fit_protein_make_specs(imol, 'all-chains') if residue_name_from_spec(imol, x) == residue_type]

# This takes 4 member specs, return True or False
#
def spec_match_qm(spec_1, spec_2):

    from types import StringType
    from types import IntType
    # first test if specs are of length 4.
    #
    if not (len(spec_1) == 4):
        return False
    if not (len(spec_2) == 4):
        return False
    chain_id_1 = spec_1[1]
    chain_id_2 = spec_2[1]
    res_no_1   = spec_1[2]
    res_no_2   = spec_2[2]
    ins_code_1 = spec_1[3]
    ins_code_2 = spec_2[3]

    if not (type(chain_id_1) is StringType):
        return False
    if not (chain_id_1 == chain_id_2):
        return False
    if not (type(res_no_1) is IntType):
        return False
    if not (res_no_1 == res_no_2):
        return False
    if not (type(ins_code_1) is StringType):
        return False
    if not (ins_code_1 == ins_code_2):
        return False
    return True
    
coot.set_console_display_commands_hilights(0, 0, 0)


#################################
# SOME MORE FUNCTIONS ONLY PYTHON
#################################

def skip_test(self, skip_query, skip_msg):
    global have_test_skip
    global skipped_tests

    if have_test_skip:
        self.skipIf(skip_query, skip_msg)
    else:
        if skip_query:
            skipped_tests.append(self.shortDescription())
            # no way so far to exit the test from here
            # either work with return value or Exceptions
            # former for now
            return True
        
unittest.TestCase.skip_test = skip_test

###############################
# NOW DEFINE THE ACTUAL TESTS
###############################


def get_this_dir():
    pass

test_file_list = ["01_pdb_mtz.py",
                  "02_shelx.py",
                  "03_ligand.py",
                  "04_cootaneer.py",
                  "05_rna_ghosts.py",
                  "06_ssm.py",
                  "07_ncs.py",
                  "08_utils.py",
                  "09_internal.py"]


# get directory of this file and execute tests found in this dir
fn = inspect.getfile(get_this_dir)
current_dir = os.path.dirname(fn)

for test_file in test_file_list:
    load_file = os.path.join(current_dir, test_file)
    load_file = os.path.normpath(load_file)
    if (os.path.isfile(load_file)):
        exec(compile(open(load_file, "rb").read(), load_file, 'exec'), globals())

test_list = [PdbMtzTestFunctions, ShelxTestFunctions,
             LigandTestFunctions, CootaneerTestFunctions,
             RnaGhostsTestFunctions, SsmTestFunctions,
             NcsTestFunctions, UtilTestFunctions,
             InternalTestFunctions]

suite = unittest.TestSuite()
for test in test_list:
    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(test))


# returns a list of all tests, in pairs containing the test class and description
# i.e. [[__main__.PdbMtzTestFunctions.test00_0, "Post Go To Atom no molecule"],
#       [__main__.PdbMtzTestFunctions.test01_0, "short discr Close bad molecule"]]
def list_of_all_tests():
    ret =[]
    for test in suite:
        ret.append([test, test.shortDescription(), test.id()])
    return ret
    
global unittest_output
unittest_output = False

# class to write output of unittest into a 'memory file' (unittest_output)
# as well as to sys.stdout
class StreamIO:
        
    def __init__(self, etxra, src=sys.stderr, dst=sys.stdout):
        import io
        global unittest_output
        unittest_output = io.StringIO()
        self.src = src
        self.dst = dst
        self.extra = unittest_output

    def write(self, msg):
        #self.src.write(msg)
        self.extra.write(msg)
        self.dst.write(msg)

    def flush(self):
        pass

class StreamIOnew:

    def __init__(self, etxra, src=sys.stderr, dst=sys.stdout):
        import io
        import io
        global unittest_output
        unittest_output = io.StringIO()
#        unittest_output = io.StringIO()
        self.src = src
        self.dst = dst
        self.extra = unittest_output

    def write(self, msg):
        #msg = msg.decode()
        self.extra.write(msg)
        self.dst.write(msg)

    def flush(self):
        pass

# function to run one test
def run_one_test(no_of_test):
    result = unittest.TextTestRunner(verbosity=2).run(list_of_all_tests()[no_of_test][0])
    return result

# function to run one test set
def run_test_set(no_of_test_set):
    print("BL DEBUG:: should run ", test_list[no_of_test_set])
    if (no_of_test_set > 0):
        read_imol_rnase_and_map()
    set = unittest.TestSuite()
    set.addTests(unittest.TestLoader().loadTestsFromTestCase(test_list[no_of_test_set]))
    result = unittest.TextTestRunner(verbosity=2).run(set)
    return result

# read imol_rnase and map if not there?!
def read_imol_rnase_and_map():
    global imol_rnase
    global imol_rnase_map

    if imol_rnase < 0:
        # not set up-> read, i.e. run test02 from 01
        imol = read_pdb(rnase_pdb())
        imol_rnase = imol

    if imol_rnase_map < 0:
        imol_map = make_and_draw_map(rnase_mtz(), "FWT","PHWT","",0,0)
        set_imol_refinement_map(imol_map)
        imol_rnase_map = imol_map
  



import numbers
# coot-lsq.py

# Copyright 2005, 2006 by Bernhard Lohkamp
# Copyright 2005, 2006 by Paul Emsley, The University of York

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,  51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA


# Internal type conversion for LSQ fitting.  Return a number
# according to the symbod match_type_in
#
def lsq_match_type_symbol(match_type_in):
  import operator

  if (isinstance(match_type_in, numbers.Number)):
     match_type_in = match_type_in
  else:
     if (match_type_in in ["CA","ca","Ca"]):
        match_type_in = 2
     elif (match_type_in in ["Main","main","mainchain","Mainchain"]):
        match_type_in = 1
     elif (match_type_in in ["ALL","all","All"]):
        match_type_in = 0
     else:
        match_type_in = -1   # unknown
  return match_type_in
#BL says, I guess I could make that more elegant...
#we use 0, 1 ,2 for ca, main, all (as it is in c++ code and not Paul's guile script)!!

# Create matchers, 7 elements:
#   [ref_start_resno, ref_end_resno, ref_chain_id, imol_ref,
#    mov_start_resno, mov_end_resno, mov_chain_id, imol_mov,
#    match_type]
#
def set_match_element(m):
# m should be a list!!!!
# something like:
# match1 = [40,50,"A",40,50,"B","all"]

    if (len(m)==7):
        match_type=lsq_match_type_symbol(m[6])
        coot.add_lsq_match(m[0],m[1],m[2],m[3],m[4],m[5],match_type)
    else:
        print("Wrong number of elements in match (was",len(m)," should be 7)")


# The scripting interface to LSQ matching.  Pass molecule numbers for
# the reference (imol_ref) and moving (imol_moving) molecules and a
# match list.  The match list format is described in the manual.
#
def lsq_match(imol_ref,imol_moving,match_list):

    coot.clear_lsq_matches()
    set_match_element(match_list)

    apply_lsq_matches(imol_ref,imol_moving)


# Simple interface to LSQ fitting.  More often than not this is what
# you will want, I imagine,
# e.g. simple_lsq_match(940, 950, "A", 0, 940, 950, "A", 1, "main")
#
def simple_lsq_match (ref_start_resno, ref_end_resno, ref_chain_id, imol_ref, mov_start_resno, mov_end_resno, mov_chain_id, imol_mov, match_type):

      internal_match_type=lsq_match_type_symbol(match_type)
      coot.clear_lsq_matches()
      coot.add_lsq_match(ref_start_resno,ref_end_resno,ref_chain_id,
		    mov_start_resno,mov_end_resno,mov_chain_id,
		    internal_match_type)
      apply_lsq_matches(imol_ref,imol_mov)


# examples:
# simple_lsq_match(940,950,"A",0,940,950,"A",1,"main")
#
# or another one:
#match1 = [40,50,"A",40,50,"B","all"]
#match2 = [45,55,"A",45,55,"B","main"]
#clear_lsq_matches()
#set_match_element(match1)
#set_match_element(match2)
#lsq_match(0,1,match1)
#
# BL says: still dont know exactly why we want that (2nd example) and what
# we should do with that! Or I havent understood and translated guile code
# wrong and screwed up!




import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk
import coot
import coot_gui
import coot_gui_api
import coot_utils
import add_linked_cho

def interactive_add_cho_dialog():

    def dummy_func():
        print("dummy_func()")
        pass

    def refine_tree_func():
        with coot_utils.UsingActiveAtom(True) as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code,
                                       aa_atom_name, aa_alt_conf, aa_res_spec]:
            refine_residues(aa_imol, glyco_tree_residues(aa_imol, aa_res_spec))

    add_linked_cho.add_synthetic_pyranose_planes()
    add_linked_cho.use_unimodal_pyranose_ring_torsions()
    # button list with [label, function]
    buttons = [
        ["Update for Current Residue", lambda _btn: dummy_func()], # dummy_func replaced later
        ["Refine Tree", lambda _btn: refine_tree_func()],
        ["Add a NAG-ASN NAG",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("NAG", "NAG-ASN")],
        ["Add a BETA1-4 NAG",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("NAG", "BETA1-4")],
        ["Add a BETA1-4 BMA",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("BMA", "BETA1-4")],
        ["Add an ALPHA1-2 MAN",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA1-2")],
        ["Add an ALPHA1-3 MAN",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA1-3")],
        ["Add an ALPHA2-3 MAN",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA2-3")],
        ["Add an ALPHA2-3 GAL",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("GAL", "ALPHA2-3")],
        ["Add an ALPHA1-6 MAN",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA1-6")],
        ["Add a BETA1-2 NAG",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("NAG", "BETA1-2")],
        ["Add a BETA1-4 GAL",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("GAL", "BETA1-4")],
        ["Add an ALPHA1-2 FUC",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("FUC", "ALPHA1-2")],
        ["Add an ALPHA1-3 FUC",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("FUC", "ALPHA1-3")],
        ["Add an ALPHA1-6 FUC",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("FUC", "ALPHA1-6")],
        ["Add an BETA1-6 FUL",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("FUL", "BETA1-6")],
        ["Add an XYP-BMA XYP",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("XYP", "XYP-BMA")],
        ["Add an ALPHA2-3 SIA",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("SIA", "ALPHA2-3")],
        ["Add an ALPHA2-6 SIA",
         lambda _btn:
         add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("SIA", "ALPHA2-6")]
    ]

    vbox = coot_gui.dialog_box_of_buttons("Add N-linked Glycan",
                                 [420, 600], buttons, "Close")[0]
    gui_add_linked_cho_dialog_vbox_set_rotation_centre_hook(vbox)
    # set the callback on the first button
    first_button = vbox.get_first_child()
    if first_button:
        first_button.connect("clicked", lambda _btn:
            gui_add_linked_cho_dialog_vbox_set_rotation_centre_hook(vbox))

    # add a widget to allow the user to choose the tree type
    #table = Gtk.Table(3, 2, False)
    table = Gtk.Grid()
    butt_1 = Gtk.CheckButton(label = "High Mannose")
    butt_2 = Gtk.CheckButton(group = butt_1, label = "Hybrid (Mammal)")
    # butt_3 = Gtk.CheckButton(group = butt_1, label = "Hybrid (Plant)")
    butt_4 = Gtk.CheckButton(group = butt_1, label = "Complex (Mammal)")
    butt_5 = Gtk.CheckButton(group = butt_1, label = "Complex (Plant)")
    butt_6 = Gtk.CheckButton(group = butt_1, label = "Expert User Mode")

    # this is now how we do it these days. FIXME-PE
    #     # add buttons for nice(?) layout/order
    #     table.attach(butt_1, 0, 1, 0, 1, Gtk.EXPAND|Gtk.FILL, Gtk.EXPAND|Gtk.FILL, 0, 0) # high mannose
    #     table.attach(butt_4, 1, 2, 0, 1, Gtk.EXPAND|Gtk.FILL, Gtk.EXPAND|Gtk.FILL, 0, 0) # complex mammal
    #     table.attach(butt_6, 0, 1, 1, 2, Gtk.EXPAND|Gtk.FILL, Gtk.EXPAND|Gtk.FILL, 0, 0) # Expert
    # #    table.attach(butt_3, 1, 2, 1, 2, Gtk.EXPAND|Gtk.FILL, Gtk.EXPAND|Gtk.FILL, 0, 0)
    #     table.attach(butt_5, 1, 2, 1, 2, Gtk.EXPAND|Gtk.FILL, Gtk.EXPAND|Gtk.FILL, 0, 0) # complex plant
    #     table.attach(butt_2, 2, 3, 0, 1, Gtk.EXPAND|Gtk.FILL, Gtk.EXPAND|Gtk.FILL, 0, 0) # hybrid mammal
    
    vbox.append(table)
    #vbox.reorder_child(table, 0)
    #vbox.reorder_child_after(table,vbox.get_first_child())

    for butt in [butt_1, butt_2, butt_4, butt_5, butt_6]:
        butt.connect("toggled", lambda func:
                     gui_add_linked_cho_dialog_vbox_set_rotation_centre_hook(vbox))

    vbox.show()

    # "global" var post-set-rotation-centre-hook
    # BL Note:: maybe should be a global!?
    global post_set_rotation_centre_script
    def post_set_rotation_centre_script():
        gui_add_linked_cho_dialog_vbox_set_rotation_centre_hook(vbox)


def glyco_tree_dialog_set_button_active_state(button, glyco_id, tree_type):

    def glyco_id2level_number(glyco_id):
        return glyco_id[0]

    def glyco_id2prime_arm_sym(glyco_id):
        return glyco_id[1]

    def glyco_id2residue_type(glyco_id):
        return glyco_id[2]

    def glyco_id2link_type(glyco_id):
        return glyco_id[3]

    def glyco_id2paren_residue_type(glyco_id):
        return glyco_id[4]

    def glyco_id2residue_spec(glyco_id):
        return glyco_id[5]
    
    def get_sensitive_button_list(glyco_id, tree_type):

        if not isinstance(glyco_id, list):
            return []
        else:
            level_number = glyco_id2level_number(glyco_id)
            prim_arm_sym = glyco_id2prime_arm_sym(glyco_id)
            residue_type = glyco_id2residue_type(glyco_id)
            link_type = glyco_id2link_type(glyco_id)
            parent_residue_type = glyco_id2paren_residue_type(glyco_id)
            residue_spec = glyco_id2residue_spec(glyco_id)
            active_button_label_ls = []
            
            if tree_type == 'expert-user-mode':
                active_button_label_ls = "expert-user-mode" # ???

            # -----------------------------------------------------------------------------------
            # Note the trees tested here match those from (get-tree-type) which examines the button
            # label of the active radio button in the dialog (these are not the auto-build trees)
            # ------------------------------------------------------------------------------------
            # BL says:: test and control the trees

            if tree_type == 'oligomannose':

                if level_number == 0:
                    if residue_type == "ASN":
                        active_button_label_ls = ["Add a NAG-ASN NAG"]
                        
                if level_number == 1:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 NAG"]
                        
                if level_number == 2:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 BMA"]

                if level_number == 3:
                    if residue_type == "BMA":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN"]

                if level_number == 4:
                    if residue_type == "MAN":
                        if link_type == "ALPHA1-3":
                            active_button_label_ls = ["Add an ALPHA1-2 MAN"]
                        if link_type == "ALPHA1-6":
                            active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                      "Add an ALPHA1-6 MAN"]
                        
                if level_number == 5:
                    if residue_type == "MAN":
                        # active_button_label_ls is the same, so no if required?
                        # or wrong tree branches. FIXME.
                        if link_type == "ALPHA1-2":
                            active_button_label_ls = ["Add an ALPHA1-2 MAN"]
                        if link_type == "ALPHA1-6":
                            active_button_label_ls = ["Add an ALPHA1-2 MAN"]
                        if link_type == "ALPHA1-3":
                            active_button_label_ls = ["Add an ALPHA1-2 MAN"]
                            
                if level_number == 6:
                    if residue_type == "MAN":
                        if link_type == "ALPHA1-2":
                            active_button_label_ls = ["Add an ALPHA1-3 GLC"]

                # inconsistencies between links here. FIXME
                if level_number == 7:
                    if residue_type == "GLC":
                        if link_type == "ALPHA1-2":
                            active_button_label_ls = ["Add an ALPHA1-3 GLC"]

                if level_number == 8:
                    if residue_type == "GLC":
                        if link_type == "ALPHA1-2":
                            active_button_label_ls = ["Add an ALPHA1-2 GLC"]

            # hybrid mammal
            if tree_type == 'hybrid-mammal':

                if level_number == 0:
                    if residue_type == "ASN":
                        active_button_label_ls = ["Add a NAG-ASN NAG"]
                        
                if level_number == 1:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 NAG",
                                                  "Add an ALPHA1-3 FUC"]
                        
                if level_number == 2:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 BMA"]

                if level_number == 2:
                    if residue_type == "FUC":
                        active_button_label_ls = ["Add a BETA1-4 GAL"]

                if level_number == 3:
                    if residue_type == "BMA":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN"]

                if level_number == 3:
                    if residue_type == "GAL":
                        active_button_label_ls = ["Add an ALPHA1-2 FUC"]

                if level_number == 4:
                    if residue_type == "MAN":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add a BETA1-2 NAG"]
                
                if level_number == 5:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 GAL"]
                
                if level_number == 5:
                    if residue_type == "GAL":
                        active_button_label_ls = ["Add an ALPHA2-3 SIA",
                                                  "Add an ALPHA2-6 SIA"]
                
            # hybrid plant
            if tree_type == 'hybrid-plant':

                if level_number == 0:
                    if residue_type == "ASN":
                        active_button_label_ls = ["Add a NAG-ASN NAG"]
                        
                if level_number == 1:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 NAG",
                                                  "Add an ALPHA1-3 FUC"]
                        
                if level_number == 2:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 BMA"]

                if level_number == 2:
                    if residue_type == "FUC":
                        active_button_label_ls = ["Add a BETA1-4 GAL"]

                if level_number == 3:
                    if residue_type == "BMA":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add an XYP-BMA XYP"]

                if level_number == 3:
                    if residue_type == "GAL":
                        active_button_label_ls = ["Add an ALPHA1-2 FUC"]

                # note that level-number is not enough for complete disambiguation,
                # we need to know we are 4 or 4' (Vliegenthart et al 1983 nomenclature).
                if level_number == 4:
                    if residue_type == "MAN":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add a BETA1-2 NAG"]
                
                if level_number == 5:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 GAL"]
                
                if level_number == 5:
                    if residue_type == "GAL":
                        active_button_label_ls = ["Add an ALPHA2-3 SIA",
                                                  "Add an ALPHA2-6 SIA"]
                
            # complex mammal
            #
            if tree_type == 'complex-mammal':

                if level_number == 0:
                    if residue_type == "ASN":
                        active_button_label_ls = ["Add a NAG-ASN NAG"]
                        
                if level_number == 1:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 NAG",
                                                  "Add an ALPHA1-6 FUC"]
                        
                if level_number == 2:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 BMA"]

                if level_number == 3:
                    if residue_type == "BMA":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add a BETA1-4 NAG"]

                #  note that level-number is not enough for complete disambiguation, 
                # we need to know we are 4 or 4' (Vl...? nomenclature).
                #
                if level_number == 4:
                    if residue_type == "MAN":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add a BETA1-2 NAG"]

                if level_number == 5:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 GAL"]

                if level_number == 6:
                    if residue_type == "GAL":
                        active_button_label_ls = ["Add an ALPHA2-3 SIA",
                                                  "Add an ALPHA2-6 SIA"]

                        
            # complex plant
            #
            if tree_type == 'complex-plant':

                if level_number == 0:
                    if residue_type == "ASN":
                        active_button_label_ls = ["Add a NAG-ASN NAG"]
                        
                if level_number == 1:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 NAG",
                                                  "Add an ALPHA1-3 FUC",
                                                  "Add an ALPHA1-6 FUC"]
                        
                if level_number == 2:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 BMA"]

                if level_number == 3:
                    if residue_type == "BMA":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add an XYP-BMA XYP",
                                                  "Add a BETA1-4 NAG"]

                if level_number == 4:
                    if residue_type == "MAN":
                        active_button_label_ls = ["Add an ALPHA1-3 MAN",
                                                  "Add an ALPHA1-6 MAN",
                                                  "Add a BETA1-2 NAG"]

                if level_number == 5:
                    if residue_type == "NAG":
                        active_button_label_ls = ["Add a BETA1-4 GAL"]

                if level_number == 6:
                    if residue_type == "GAL":
                        active_button_label_ls = ["Add an ALPHA2-3 SIA",
                                                  "Add an ALPHA2-6 SIA"]
                        
            return active_button_label_ls

    # main line
    #
    l = button.get_label()
    active_button_label_ls = get_sensitive_button_list(glyco_id, tree_type)
    if (active_button_label_ls == "expert-user-mode"):
        button.set_sensitive(True)
    else:
        if not (l == "Update for Current Residue") and not (l == "Refine Tree"):
            button.set_sensitive(l in active_button_label_ls)

def gui_add_linked_cho_dialog_vbox_set_rotation_centre_hook(vbox):
    """
    vbox is the vbox of the dialog box of buttons. One of the children of the vbox
    is the table that contains the buttons
    """

    def get_tree_type():
        tree_type = "oligomannose"
        child = vbox.get_first_child()
        def iterate():
            child = vbox.get_next_sibling()
            return child is not None
        while iterate():
            if type(child) == Gtk.Table:
                for table_child in child:
                    if type(table_child) == Gtk.CheckButton:
                        if table_child.get_active():
                            l = table_child.get_label()
                            # this is a bit ugly because we are testing that these strings
                            # match button labels (set in interactive-add-cho-dialog)
                            if l == "High Mannose":
                                tree_type = 'oligomannose'
                            if l == "Hybrid (Mammal)":
                                tree_type = 'hybrid-mammal'
                            if l == "Hybrid (Plant)":
                                tree_type = 'hybrid-plant'
                            if l == "Expert User Mode":
                                tree_type = 'expert-user-mode'
                            if l == "Complex (Mammal)":
                                tree_type = 'complex-mammal'
                            if l == "Complex (Plant)":
                                tree_type = 'complex-plant'
        return tree_type

    with coot_utils.UsingActiveAtom(True) as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code,
                                   aa_atom_name, aa_alt_conf, aa_res_spec]:
        glyco_id = coot.glyco_tree_residue_id_py(aa_imol, aa_res_spec)
        # Paule says:
        # if it was an ASP create a level-0 glyco-id for that (glyco-tree-residue-id doesn't
        # do that (not sure why)).
        if not glyco_id:
            rn = coot.residue_name(aa_imol, aa_chain_id, aa_res_no, aa_ins_code)
            if isinstance(rn, str):
                if rn == "ASN":
                    glyco_id = [0, "unset", "ASN", "", "", aa_res_spec]
        if isinstance(glyco_id, list):
            tree_type = get_tree_type()
            for child in vbox:
                if (type(child) == Gtk.Button):
                    glyco_tree_dialog_set_button_active_state(child, glyco_id,
                                                              tree_type)
            return True
        return False


# There is some graphics to the validation here, so keep it in the graphics part
# wont be missed in the non-grpahical part for now, I think (BL that is).
    
# this has to be at the top level for reasons that I don't understand
# (python function scoping) - I'll just put it down to python being shit as usual.
#
def glyco_validation_dialog_set_go_to_residue(imol, residue_spec):
    rc = residue_centre(imol,
                        res_spec_utils.residue_spec_to_chain_id(residue_spec),
                        res_spec_utils.residue_spec_to_res_no(residue_spec),
                        '')
    coot.set_rotation_centre(*rc)

def load_privateer_dictionary():
    if os.path.exists("privateer-lib.cif"):
        coot.read_cif_dictionary("privateer-lib.cif")
        coot.set_refine_with_torsion_restraints(1)
    
class glyco_validate:

    def run_privateer(self, imol, glyco_tree_residues, hklin_fn, fp_col, sigfp_col, pdbin_fn, privateer_log):
        if coot.is_valid_model_molecule(imol):
            args = ['-mtzin', hklin_fn, '-colin-fo', fp_col+','+sigfp_col,
                    '-pdbin', pdbin_fn]
            # using'-mode', 'ccp4i2' makes a file with no residue nambers (afaics)
            coot_utils.popen_command("privateer", args, [], privateer_log, False)

    def make_privateer_validation_info(self, imol, fp_col, sigfp_col, glyco_tree_residues):

        imol_map = coot.imol_refinement_map()

        if len(glyco_tree_residues) > 0:
            d = coot_utils.get_directory("coot-ccp4")
            spid = str(os.getpid())
            fn_pdb = "coot-privateer-" + spid + ".pdb"
            fn_log = "coot-privateer-" + spid + ".log"
            privateer_pdb = os.path.join(d, fn_pdb)
            privateer_log = os.path.join(d, fn_log)
            hklin_fn = coot.mtz_file_name(imol_map);
            coot.write_pdb_file(imol, privateer_pdb)
            self.run_privateer(imol, glyco_tree_residues, hklin_fn, fp_col, sigfp_col, privateer_pdb, privateer_log)
            pvi = self.parse_privateer_log(privateer_log, imol, glyco_tree_residues)
            return pvi
        else:
            return []

    # return a list of residues with privateer validation info
    #
    def parse_privateer_log(self, log_file_name, imol, glyco_tree_residues):

        print('parse_privateer_log', log_file_name, imol, glyco_tree_residues)

        pvi = []
        f = open(log_file_name)
        lines = f.readlines()
        f.close()
        for line in lines:
            l = line.rstrip()
            if len(l) > 10:
                if l[:4] == 'coot':
                    words = l.split()
                    if len(words) > 12:
                        for r in glyco_tree_residues:
                            rn = residue_name_by_spec(imol, r)
                            try:
                                res_id = rn + "-" + res_spec_utils.residue_spec_to_chain_id(r) + \
                                         '-' + str(coot_utils.residue_spec_to_res_no(r))
                                # print "res_id", res_id
                                if words[1] == res_id:
                                    # print words[12] , yes or check
                                    new_item = (r, words)
                                    pvi.append(new_item)
                            except TypeError as e:
                                print(e)
        print("parsed", log_file_name)
        return pvi

    def make_validation_dialog(self, imol, privateer_validation_info):

        print("make_validation_dialog", privateer_validation_info)
        buttons = []
        for vi in privateer_validation_info:
            residue_spec = vi[0]
            words = vi[1]
            state = words[12]
            if state == "yes":
                state = " OK  " # label spacing
            button_text = words[1] + "   Q=" + words[3] + " RSCC=" + words[6] + \
                          " Cnf=" + words[8] + "   " + state
            # string->function? Hideous
            func = "glyco_validation_dialog_set_go_to_residue(" + str(imol) + "," + \
                   str(residue_spec) + ")"
            button = [button_text, func]
            buttons.append(button)
        if len(buttons) > 0:
            button = ["Load Privateer Dictionary", "load_privateer_dictionary()"]
            buttons.append(button)
            coot_gui.dialog_box_of_buttons("Privateer Validation", (400, 220), buttons, " Close ")

    def validation_dialog(self):
        
        active_atom = coot.active_residue_py()
        try:
            imol = active_atom[0]
            active_residue = active_atom[:4]
            glyco_tree_residues = coot.glyco_tree_residues_py(imol, active_residue)

            print('imol', imol)
            print('active_residue', active_residue)
            print('glyco_tree_residues', glyco_tree_residues)

            fp_col='FP'
            sigfp_col='SIGFP'    
            pvi = self.make_privateer_validation_info(imol, fp_col, sigfp_col, glyco_tree_residues)
            # print "validation_dialog()", pvi
            if len(pvi) > 0:
                # now make a gui
                self.make_validation_dialog(imol, pvi)

        except KeyError as e:
            print(e)

        # no active atom
        except TypeError as e:
            print(e)

    def auto_delete_residues_internal(self, imol, glyco_tree_residues):
        fp_col='FP'
        sigfp_col='SIGFP'
        pvi = self.make_privateer_validation_info(imol, fp_col, sigfp_col, glyco_tree_residues)
        for res_info in pvi:
            print("test", res_info)
            res_status = res_info[1][12]
            if res_status == 'check' or res_status == 'no':
                coot_utils.delete_residue_by_spec(imol, res_info[0])

    def auto_delete_residues(self):
        try:
            active_atom = coot.active_residue_py()
            imol = active_atom[0]
            active_residue = active_atom[:4]
            glyco_tree_residues = coot.glyco_tree_residues_py(imol, active_residue)
            self.auto_delete_residues_internal(imol, glyco_tree_residues)
        except TypeError as e:
            print(e)
    
            
# graphics...

def add_module_carbohydrate_gui():
    menu = coot_gui.attach_module_menu_button("Glyco")

    coot_gui.add_simple_action_to_menu(
        menu, "N-linked Glycan Addition...","interactive_add_cho_dialog",
        lambda _simple_action, _two:
        interactive_add_cho_dialog())
    
    def add_multi_carbo_link_func(link_list):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, aa_atom_name, aa_alt_conf]:
            add_linked_cho.multi_add_linked_residue(aa_imol,
                                        [aa_chain_id, aa_res_no, aa_ins_code],
                                        link_list)


    def set_default_cho_b_factor_func():
        with coot_utils.UsingActiveAtom(True) as [aa_imol, aa_chain_id, aa_res_no,
                                        aa_ins_code, aa_atom_name,
                                        aa_alt_conf, aa_res_spec]:
            residues = coot.residues_near_residue_py(aa_imol, aa_res_spec, 10)
            imol_region = coot.new_molecule_by_residue_specs(aa_imol, residues)
            # BL says:: why not do a new mol by sphere selection?!
            m = coot.median_temperature_factor(imol_region)
            coot.close_molecule(imol_region)
            if coot_utils.isNumber(m):
                new_m = m* 1.55
                coot.set_default_temperature_factor_for_new_atoms(new_m)
                s = "New Temperature Factor set to " + str(new_m)
                coot.info_dialog(s)
        
    coot_gui.add_simple_action_to_menu(
        menu, "Set Default N-linked CHO Atoms B-factor","set_default_cho_b_factor",
        lambda _simple_action, _two: set_default_cho_b_factor_func()
    )

    
    coot_gui.add_simple_action_to_menu(
        menu, "N-link add NAG, NAG, BMA","multi_carbo_link_NAG_NAG_BMA",
        lambda _simple_action, _two: add_multi_carbo_link_func([["NAG", "NAG-ASN"],
                                                ["NAG", "BETA1-4"],
                                                ["BMA", "BETA1-4"]]))
    
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add a ASN-NAG NAG",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("NAG", "NAG-ASN"))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add a BETA1-4 NAG",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("NAG", "BETA1-4"))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add a BETA1-4 BMA",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("BMA", "BETA1-4"))
    
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA1-2 MAN",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA1-2"))
    
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA1-3 MAN",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA1-3"))

    # we should do this only if we are sitting on an SIA.
    # Attaching a SIA to a MAN (i.e. reverse order) would be a
    # good test too...
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA2-3 MAN",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA2-3"))

    # # same consideration as above
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA2-3 GAL",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("GAL", "ALPHA2-3"))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA1-6 MAN",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("MAN", "ALPHA1-6"))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA1-3 FUC",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("FUC", "ALPHA1-3"))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an ALPHA1-6 FUC",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("FUC", "ALPHA1-6"))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Add an XYP-BMA XYP",
    #     lambda _simple_action, _two:
    #     add_linked_cho.add_linked_residue_with_extra_restraints_to_active_residue("XYP", "XYP-BMA"))


    # the mode in the function call now takes take of this
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Auto Fit & Refine On for Link Addition",
    #     lambda _simple_action, _two: coot.set_add_linked_residue_do_fit_and_refine(1))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Auto Fit & Refine Off for Link Addition",
    #     lambda _simple_action, _two: coot.set_add_linked_residue_do_fit_and_refine(0))

    def add_oligo_tree_func(oligo_tree):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, aa_atom_name, aa_alt_conf]:
            coot.make_backup(aa_imol)
            # switch backup off?!
            add_linked_cho.add_linked_residue_tree(aa_imol,
                                    [aa_chain_id, aa_res_no, aa_ins_code],
                                    oligo_tree)
        
    coot_gui.add_simple_action_to_menu(
        menu, "Add High Mannose","add_high_mannose",
        lambda _simple_action, _two: add_oligo_tree_func(add_linked_cho.oligomannose_tree()))

    coot_gui.add_simple_action_to_menu(
        menu, "Add Hybrid (Mammal)","add_hybrid_mammal",
        lambda _simple_action, _two: add_oligo_tree_func(add_linked_cho.hybrid_mammal_tree()))
    
#            in practice, no one will be doing this. 
#            coot_gui.add_simple_action_to_menu(
#                menu, "Add Hybrid (Plant)",
#                lambda _simple_action, _two: add_oligo_tree_func(add_linked_cho.hybrid_plant_derived_tree()))

    coot_gui.add_simple_action_to_menu(
        menu, "Add Complex (Mammal)","add_complex_mammal",
        lambda _simple_action, _two: add_oligo_tree_func(add_linked_cho.complex_mammal_tree()))

    coot_gui.add_simple_action_to_menu(
        menu, "Add Complex (Plant)","add_complex_plant",
        lambda _simple_action, _two: add_oligo_tree_func(add_linked_cho.complex_plant_tree()))

    coot_gui.add_simple_action_to_menu(
        menu, "Delete All Carbohydrate","delete_all_cho",
        lambda _simple_action, _two: add_linked_cho.delete_all_cho())

    def torsion_fit_this_func(refine = False):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, aa_atom_name, aa_alt_conf]:
            centre_residue = [aa_chain_id,aa_res_no, aa_ins_code]
            coot.multi_residue_torsion_fit(aa_imol,
                                        [centre_residue],
                                        30000)
            if refine:
                with AutoAccept():
                    refine_residues(aa_imol, [centre_residue])

    def torsion_fit_this_and_neighbours_func(refine = False):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, aa_atom_name, aa_alt_conf]:
            centre_residue = [aa_chain_id,aa_res_no, aa_ins_code]
            residues = coot.residues_near_residue_py(aa_imol, centre_residue, 1.9)
            residues.append(centre_residue)
            coot.multi_residue_torsion_fit(aa_imol, residues, 30000)
            if refine:
                with AutoAccept():
                    refine_residues(aa_imol, [centre_residue])

    coot_gui.add_simple_action_to_menu(
        menu, "Torsion Fit this residue","torsion_fit_this",
        lambda _simple_action, _two: torsion_fit_this_func())

    # coot_gui.add_simple_action_to_menu(
    #     menu, "Torsion Fit This Residue and Neighbours",
    #     lambda _simple_action, _two: torsion_fit_this_and_neighbours_func())

    coot_gui.add_simple_action_to_menu(
        menu, "Torsion Fit & Refine this residue","torsion_fit_refine_this",
        lambda _simple_action, _two: torsion_fit_this_func(True))

    coot_gui.add_simple_action_to_menu(
        menu, "Add synthetic pyranose plane restraints","add_synthetic_pyranose_planes",
        lambda _simple_action, _two: add_linked_cho.add_synthetic_pyranose_planes())

    coot_gui.add_simple_action_to_menu(
        menu, "Use Unimodal ring torsion restraints","use_unimodal_pyranose_ring_torsions",
        lambda _simple_action, _two: add_linked_cho.use_unimodal_pyranose_ring_torsions())

    # This should probably become a checkbox??
    # note: this is duplicated in Restraints

    coot_gui.add_simple_action_to_menu(
        menu, "Display Extra Restraints","cho__display_extra_restraints",
        lambda _simple_action, _two: coot_utils.using_active_atom(set_show_extra_restraints, "aa_imol", 1))

    coot_gui.add_simple_action_to_menu(
        menu, "Undisplay Extra Restraints","cho__undisplay_extra_restraints",
        lambda _simple_action, _two: coot_utils.using_active_atom(set_show_extra_restraints, "aa_imol", 0))

    coot_gui.add_simple_action_to_menu(
        menu, "Extract this Tree","new_molecule_from_this_glyco_tree",
        lambda _simple_action, _two:
        add_linked_cho.new_molecule_from_this_glyco_tree())



cprodrg = "cprodrg"
# we cannot use full path as cprodrg is spawning refmac...
#cprodrg = "c:/Programs/CCP4-Packages/ccp4-6.1.13/bin/cprodrg.exe"

# we need $CLIBD to run prodrg (for prodrg.param), so check for it:
if not os.getenv("CLIBD"):
    # print "BL INFO:: potential problem to run prodrg (no CLIBD set)"
    # print "          try to fix this"
    bin_dir = os.path.dirname(cprodrg)
    base_dir = os.path.dirname(bin_dir)
    clibd = os.path.join(base_dir, "lib", "data")
    prodrg_params = os.path.join(clibd, "prodrg.param")
    if os.path.isfile(prodrg_params):
        os.environ["CLIBD"] = clibd
        print "BL INFO:: found prodrg.param"
        print "          and set CLIBD to", clibd
    else:
        # print "BL ERROR:: no prodrg.param available"
        # print "searched in ", clibd
        if coot.enhanced_ligand_coot_p():
            # we don't want to know that you can't find old programs
            pass
        else:
           print "Sorry, prodrg module not available"
        # should load rest?
        # i.e. stop here?
    

# these are the files that mdl-latest-time and sbase time-out functions
# look at.
# 
# if there is a prodrg-xyzin set the current-time to its mtime, else False
#
global prodrg_xyzin
global sbase_to_coot_tlc
prodrg_xyzin      = "prodrg-in.mdl"
sbase_to_coot_tlc = ".sbase-to-coot-comp-id"

# this is for BL win machine
home = os.getenv("HOME")
if (not home and coot_utils.is_windows()):
    home = os.getenv("COOT_HOME")
if home:
    prodrg_xyzin      = os.path.join(home, "Projects",
                                     "build-xp-python", "lbg", "prodrg-in.mdl")
else:
    print "BL DEBUG:: buggery home is", home  # FIXME
sbase_to_coot_tlc = "../../build-xp-python/lbg/.sbase-to-coot-comp-id"

# this is latest!!!!
prodrg_xyzin      = "prodrg-in.mdl"
sbase_to_coot_tlc = ".sbase-to-coot-comp-id"

# to be over-ridden by your favourite 3d conformer generator, if you like...
#
def import_from_3d_generator_from_mdl(file_name, comp_id):
    import_from_prodrg("mini-no", comp_id)


def import_ligand_with_overlay(prodrg_xyzout, prodrg_cif):
    
    # OK, so here we read the PRODRG files and
    # manipulate them.  We presume that the active
    # residue is quite like the input ligand from
    # prodrg.
    #
    # Read in the lib and coord output of PRODRG.  Then
    # overalay the new ligand onto the active residue
    # (just so that we can see it approximately
    # oriented). Then match the torsions from the new
    # ligand to the those of the active residue.  Then
    # overlay again so that we have the best match.
    #
    # We want to see just one molecule with the protein
    # and the new ligand.
    # add_ligand_delete_residue_copy_molecule provides
    # that for us.  We just colour it and undisplay the
    # other molecules.
    
    # overlap the imol_ligand residue if there are restraints for the
    # reference residue/ligand.
    #
    # Don't overlap if the reference residue/ligand is not a het-group.
    #
    # return overlapped status
    #
    def overlap_ligands_maybe(imol_ligand, imol_ref, chain_id_ref, res_no_ref):

        # we don't want to overlap-ligands if there is no dictionary
        # for the residue to be matched to.
        res_name = coot.residue_name(imol_ref, chain_id_ref, res_no_ref, "")
        restraints = monomer_restraints(res_name)
        if (not restraints):
            return False
        else:
            if not coot_utils.residue_has_hetatms_qm(imol_ref, chain_id_ref, res_no_ref, ""):
                return False
            else:
                print "----------- overlap-ligands %s %s %s %s ------------" \
                      %(imol_ligand, imol_ref, chain_id_ref, res_no_ref)
                # this can return the rtop operator or False (for fail of course).
                return overlap_ligands(imol_ligand, imol_ref, chain_id_ref, res_no_ref)
            
    # return the new molecule number.
    #
    def read_and_regularize(prodrg_xyzout):
        imol = coot.handle_read_draw_molecule_and_move_molecule_here(prodrg_xyzout)
        # speed up the minisation (and then restore setting).
        # No need to put the refinement step stuff in auto accept!?
        s = coot.dragged_refinement_steps_per_frame()
        coot.set_dragged_refinement_steps_per_frame(500)
        with AutoAccept():
            coot.regularize_residues( imol, [["", 1, ""]])
        coot.set_dragged_refinement_steps_per_frame(s)
        return imol

    # return the new molecule number
    # (only works with aa_ins_code of "")
    #
    def read_regularize_and_match_torsions_maybe(prodrg_xyzout, imol_ref,
                                                 chain_id_ref, res_no_ref):
        imol = coot.handle_read_draw_molecule_and_move_molecule_here(prodrg_xyzout)

        if (not have_restraints_for_qm(coot.residue_name(imol_ref, chain_id_ref,
                                                    res_no_ref, ""))):
            return False
        else:
            overlap_status = overlap_ligands_maybe(imol, imol_ref,
                                                   chain_id_ref, res_no_ref)
            
            # speed up the minisation (and then restore setting).
            # No need to put refinement steps in auto accept!?
            s = coot.dragged_refinement_steps_per_frame()
            coot.set_dragged_refinement_steps_per_frame(600)
            with AutoAccept():
                coot.regularize_residues(imol, [["", 1, ""]])
            coot.set_dragged_refinement_steps_per_frame(s)
            if overlap_status:
                coot.match_ligand_torsions(imol, imol_ref, chain_id_ref, res_no_ref)
        return imol

    # return True or False
    #
    def have_restraints_for_qm(res_name):
        restraints = monomer_restraints(res_name)
        return not restraints == False   # twisted but short

    # main line

    coot.read_cif_dictionary(prodrg_cif)

    # we do different things depending on whether or
    # not there is an active residue.  We need to test
    # for having an active residue here (currently we
    # presume that there is).
    # 
    # Similarly, if the aa_ins_code is non-null, let's
    # presume that we do not have an active residue.

    active_atom = active_residue()
    if ((not active_atom) or
        (not active_atom[3] == "")):  # aa_ins_code

        # then there is no active residue to match to
        read_and_regularize(prodrg_xyzout)
        # BL says:: maybe we should merge the ligand into
        # the protein molecule?! But what is the protein?
        # merge_molecules([active_atom[0]], imol)
    else:
        # we have an active residue to match to
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code,
                                   aa_atom_name, aa_alt_conf]:
            if not coot_utils.residue_is_close_to_screen_centre_qm(
                aa_imol, aa_chain_id, aa_res_no, ""):

                # not close, no overlap
                #
                read_and_regularize(prodrg_xyzout)
            else:
                # try overlap
                # BL says:: this overwrites 
                imol = read_regularize_and_match_torsions_maybe(prodrg_xyzout,
                                                                aa_imol, aa_chain_id, aa_res_no)

                overlapped_flag = overlap_ligands_maybe(imol, aa_imol,
                                                        aa_chain_id, aa_res_no)

                if overlapped_flag:
                    print "------ overlapped-flag was true!!!!!"
                    coot.set_mol_displayed(aa_imol, 0)
                    coot.set_mol_active(aa_imol, 0)
                    col = coot.get_molecule_bonds_colour_map_rotation(aa_imol)
                    new_col = col + 5 # a tiny amount
                    # new ligand specs, then "reference" ligand (to be deleted)
                    imol_replaced = coot.add_ligand_delete_residue_copy_molecule(
                        imol, "", 1,
                        aa_imol, aa_chain_id, aa_res_no)


                    coot.set_molecule_bonds_colour_map_rotation(imol_replaced, new_col)
                    coot.set_mol_displayed(imol, 0)
                    coot.set_mol_active(imol, 0)
                    coot.graphics_draw()

    return True
    # return False otherwise? When? FIXME

    
def import_from_prodrg(minimize_mode, res_name):

    import operator
    global prodrg_xyzin

    # main line of import_from_prodrg
    #
    prodrg_dir = "coot-ccp4"

    coot.make_directory_maybe(prodrg_dir)
    prodrg_xyzout = os.path.join(prodrg_dir, "prodrg-" + res_name + ".pdb")
    prodrg_cif    = os.path.join(prodrg_dir, "prodrg-out.cif")
    prodrg_log    = os.path.join(prodrg_dir, "prodrg.log")

    # FIXME? maybe, we could use 'enum' for mini-no rather than string?!
    # see news_status stuff
    # requires python >= 2.5 (shall we test?)
    mini_mode = "NO" if (minimize_mode == 'mini-no') else "PREP"
    # see if we have cprodrg
    if not (os.path.isfile(cprodrg) or
            coot_utils.command_in_path_qm(cprodrg)):
        coot.info_dialog("BL INFO:: No cprodrg found")
    else:
        status = coot_utils.popen_command(cprodrg,
                               ["XYZIN",  prodrg_xyzin,
                                "XYZOUT", prodrg_xyzout,
                                "LIBOUT", prodrg_cif],
                               ["MINI " + mini_mode, "END"],
                               prodrg_log, True)
        if isinstance(status, int):
            if (status == 0):
                import_ligand_with_overlay(prodrg_xyzout, prodrg_cif)


def get_file_latest_time(file_name):
    if not os.path.isfile(file_name):
        return False
    else:
        return os.stat(file_name).st_mtime

# globals should be in the beginning! FIXME
global mdl_latest_time
global sbase_transfer_latest_time
mdl_latest_time = get_file_latest_time(prodrg_xyzin)
sbase_transfer_latest_time = get_file_latest_time(sbase_to_coot_tlc)
# FIXME: this is not a proper name
def mdl_update_timeout_func():

    import operator
    global mdl_latest_time
    global sbase_transfer_latest_time
    
    mdl_now_time   = get_file_latest_time(prodrg_xyzin)
    sbase_now_time = get_file_latest_time(sbase_to_coot_tlc)

    # print "sbase_now_time %s    sbase_latest_time %s" %(sbase_now_time, sbase_transfer_latest_time)

    if (operator.isNumberType(mdl_now_time)):
        if operator.isNumberType(mdl_latest_time):
            if (mdl_now_time > mdl_latest_time):
                mdl_latest_time = mdl_now_time
                import_from_prodrg('mini-prep')

    if operator.isNumberType(sbase_transfer_latest_time):
        if operator.isNumberType(sbase_now_time):
            if (sbase_now_time > sbase_transfer_latest_time):
                sbase_transfer_latest_time = sbase_now_time
                try: # sort of check if file exists?
                    fin = open(sbase_to_coot_tlc, 'r')
                    tlc_symbol = fin.readline()  # need to read more? FIXME
                    fin.close()
                    imol = coot.get_ccp4srs_monomer_and_dictionary(tlc_symbol)
                    if not coot_utils.valid_model_molecule_qm(imol):
                        print "failed to get SBase molecule for", tlc_symbol
                    else:
                        # it was read OK, do an overlap
                        coot_utils.using_active_atom(overlap_ligands, imol,
                                          "aa_imol", "aa_chain_id", "aa_res_no")
                except:
                    print "BL ERROR:: reading sbase file", sbase_to_coot_tlc
                
    return True # return value, keep running; FIXME:: how to stop?


# This is done internally now, by passing lbg a function that runs
# prodrg after the mdl file has been written.  On demand.  We no
# longer need to have a timeout looking for a new prodrg-in.mdl file.
#
# gobject.timeout_add(500, mdl_update_timeout_func)

# return False (if fail) or a list of: the molecule number of the
# selected residue, the prodrg output mol file_name, the prodrg
# output pdb file_name
#
def prodrg_flat(imol_in, chain_id_in, res_no_in):
    """return False (if fail) or a list of: the molecule number of the
    selected residue, the prodrg output mol file_name, the prodrg
    output pdb file_name

    Keyword arguments:
    imol_in -- the molecule number
    chain_id_in -- chain id of the molecule
    res_no_in -- residue number ot the molecule
    
    """

    import operator
    import random

    # return a text string, or at least "" if we can't find the prodrg
    # error message output line.
    def get_prodrg_error_message(log_file):
        if not os.path.isfile(log_file):
            return ""
        else:
            fin = open(log_file, 'r')
            lines = fin.readlines()
            fin.close()
            for line in lines:
                if (" PRODRG: " in line):
                    # assuming only one error?!
                    return line
            return ""  # no error

    selection_string = "//" + chain_id_in + "/" + str(res_no_in)
    imol = coot.new_molecule_by_atom_selection(imol_in, selection_string)
    prodrg_input_file_name = os.path.join("coot-ccp4", "tmp-residue-for-prodrg.pdb")
    prodrg_output_mol_file = os.path.join("coot-ccp4", ".coot-to-lbg-mol")
    prodrg_output_pdb_file = os.path.join("coot-ccp4", ".coot-to-lbg-pdb")
    prodrg_output_lib_file = os.path.join("coot-ccp4", ".coot-to-lbg-lib")
    prodrg_log             = os.path.join("coot-ccp4", "tmp-prodrg-flat.log")
    coot.set_mol_displayed(imol, 0)
    set_mol_active   (imol, 0)
    coot.write_pdb_file(imol, prodrg_input_file_name)
    coot.make_directory_maybe("coot-ccp4")
    arg_list = ["XYZIN",  prodrg_input_file_name,
                "MOLOUT", prodrg_output_mol_file,
                "XYZOUT", prodrg_output_pdb_file,
                "LIBOUT", prodrg_output_lib_file]
    print "arg_list", arg_list
    status = coot_utils.popen_command(cprodrg,
                           arg_list,
                           ["COORDS BOTH", "MINI FLAT", "END"],
                           prodrg_log, True)
    # Does this make sense in python? status being number that is.
    # Maybe rather check for cprodrg exe. FIXME
    if not operator.isNumberType(status):
        coot.info_dialog("Ooops: cprodrg not found?")
        return False
    else:
        if not (status == 0):
            # only for python >=2.5
            mess = "Something went wrong running cprodrg\n" + \
                   get_prodrg_error_message(prodrg_log)
            coot.info_dialog(mess)
            return False
        else:
            # normal return value (hopefully)
            return [imol,
                    prodrg_output_mol_file,
                    prodrg_output_pdb_file,
                    prodrg_output_lib_file]


def prodrg_plain(mode, imol_in, chain_id_in, res_no_in):

    selection_string = "//" + chain_id_in + "/" + \
                       str(res_no_in)
    imol = coot.new_molecule_by_atom_selection(imol_in, selection_string)
    stub = os.path.join("coot-ccp4", "prodrg-tmp-" + str(os.getpid()))
    prodrg_xyzin  = stub + "-xyzin.pdb"
    prodrg_xyzout = stub + "-xyzout.pdb"
    prodrg_cif    = stub + "-dict.cif"
    prodrg_log    = stub + ".log"

    coot.write_pdb_file(imol, prodrg_xyzin)
    result = coot_utils.popen_command(cprodrg,
                           ["XYZIN",  prodrg_xyzin,
                            "XYZOUT", prodrg_xyzout,
                            "LIBOUT", prodrg_cif],
                           ["MINI PREP", "END"],
                           prodrg_log, True)
    coot.close_molecule(imol)
    return [result, prodrg_xyzout, prodrg_cif]


# Why do we pass imol etc and then use active atom?
def fle_view(imol, chain_id, res_no, ins_code):

    import operator
    global have_mingw
    
    # not using active atom, but make property list
    r_flat  = prodrg_flat (imol, chain_id, res_no)
    r_plain = prodrg_plain('mini-no', imol, chain_id, res_no)
    if (r_flat and (r_plain[0] == 0 )):
        imol_ligand_fragment = r_flat[0]
        prodrg_output_flat_mol_file_name = r_flat[1]
        prodrg_output_flat_pdb_file_name = r_flat[2]
        prodrg_output_cif_file_name      = r_flat[3]
        prodrg_output_3d_pdb_file_name   = r_plain[1]
        # 'using_active_atom'
        active_atom = active_residue()
        aa_imol     = active_atom[0]
        aa_chain_id = active_atom[1]
        aa_res_no   = active_atom[2]
        coot.fle_view_internal(aa_imol, aa_chain_id, aa_res_no, "",  # should be from active_atom!!     coot_utils.using_active_atom([[]])
                          imol_ligand_fragment,
                          prodrg_output_flat_mol_file_name,
                          prodrg_output_flat_pdb_file_name,
                          prodrg_output_3d_pdb_file_name,
                          prodrg_output_cif_file_name)
        # touch on Windows!?
        # either distribute touch.exe or use DOS:
        # for non existing file use:
        # copy NUL YourFile.txt
        # for existing:
        # copy /b filename.ext +,,
        #
        # all not neede any more...
        # BL says:: just keep in case I need to touch anything again!!!
        #
        #if (coot_utils.is_windows()):
        #    import subprocess
        #    lbg_ready = os.path.abspath(os.path.join("coot-ccp4",
        #                                              ".coot-to-lbg-mol-ready"))
        #    print "BL DEBUG:: lbg_ready is", lbg_ready
        #    if os.path.isfile(lbg_ready):
        #        subprocess.call("copy /b /y " + lbg_ready + " +,, " + lbg_ready
        #                        , shell=True)
        #    else:
        #        subprocess.call("copy NUL " + lbg_ready, shell=True)
        #else:
        #    coot_utils.popen_command("touch",
        #                  [os.path.join("coot-ccp4",
        #                                ".coot-to-lbg-mol-ready")],
        #                  [],
        #                  "/dev/null", False)

# using cprodrg
# again, why passing imol etc? FIXME
#
def fle_view_to_png(imol, chain_id, res_no, ins_code, neighb_radius,
                    png_file_name):

    # not using active atom, but make property list
    # 'using_active_atom'
    active_atom = active_residue()
    imol     = active_atom[0]
    chain_id = active_atom[1]
    res_no   = active_atom[2]
    
    r_flat  = prodrg_flat (imol, chain_id, res_no)
    r_plain = prodrg_plain('mini-no', imol, chain_id, res_no)
    
    if (r_flat and (r_plain[0] == 0 )):
        imol_ligand_fragment = r_flat[0]
        prodrg_output_flat_mol_file_name = r_flat[1]
        prodrg_output_flat_pdb_file_name = r_flat[2]
        prodrg_output_cif_file_name      = r_flat[3]
        prodrg_output_3d_pdb_file_name   = r_plain[1]
        coot.fle_view_internal_to_png(imol, chain_id, res_no, "",  # should be from active_atom!!     coot_utils.using_active_atom([[]])
                                 imol_ligand_fragment,
                                 prodrg_output_flat_mol_file_name,
                                 prodrg_output_flat_pdb_file_name,
                                 prodrg_output_3d_pdb_file_name,
                                 prodrg_output_cif_file_name, 1,
                                 png_file_name)
                

# import from SRS, callback using sbase_import_function
#
def get_ccp4srs_monomer_and_overlay(comp_id):

    """
    import from SBASE, callback using sbase_import_function
    """

    if (active_residue()):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                   aa_ins_code, aa_atom_name, aa_alt_conf]:
            imol = coot.get_ccp4srs_monomer_and_dictionary(comp_id)
            overlap_ligands(imol, aa_imol, aa_chain_id, aa_res_no)
            # BL:: again? I dont think so
        coot.get_ccp4srs_monomer_and_dictionary(comp_id)




add_key_binding("Refine Active Residue", "r", lambda: fitting.manual_refine_residues(0))
add_key_binding("Refine Active Residue AA", "x", lambda: fitting.refine_active_residue())
add_key_binding("Triple Refine", "t", lambda: fitting.manual_refine_residues(1))
add_key_binding("Autofit Rotamer", "j", lambda: fitting.auto_fit_rotamer_active_residue())
add_key_binding("Pepflip", "q", lambda: fitting.pepflip_active_residue())
add_key_binding("Go To Blob", "g", lambda: coot.blob_under_pointer_to_screen_centre())
add_key_binding("Eigen-flip Ligand", "e", lambda: coot_utils.flip_active_ligand())
add_key_binding("Add Water", "w", lambda: coot.place_typed_atom_at_pointer("Water"))

def add_water_in_blob():
    coot.blob_under_pointer_to_screen_centre()
    coot.place_typed_atom_at_pointer("Water")
    fitting.refine_active_residue()
add_key_binding("Add Water + (centre+refine)", "W", lambda: add_water_in_blob())

def key_binding_func_1():
    active_atom = active_residue()
    if (not active_atom):
        print("No active atom")
    else:
        imol      = active_atom[0]
        chain_id  = active_atom[1]
        res_no    = active_atom[2]
        ins_code  = active_atom[3]
        atom_name = active_atom[4]
        alt_conf  = active_atom[5]
        coot.add_terminal_residue(imol, chain_id, res_no, "auto", 1)
add_key_binding("Add terminal residue", "y", lambda: key_binding_func_1())

def key_binding_func_2():
    active_atom = active_residue()
    if (not active_atom):
        print("No active atom")
    else:
        imol      = active_atom[0]
        chain_id  = active_atom[1]
        res_no    = active_atom[2]
        ins_code  = active_atom[3]
        atom_name = active_atom[4]
        alt_conf  = active_atom[5]
        coot.fill_partial_residue(imol, chain_id, res_no, ins_code)
add_key_binding("Fill Partial", "k", lambda: key_binding_func_2())

def delete_residue_sidechain_key():
    with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                               aa_ins_code, aa_atom_name, aa_alt_conf]:
        coot.delete_residue_sidechain(aa_imol, aa_chain_id, aa_res_no,
                                 aa_ins_code, 0)
add_key_binding("Delete Residue Sidechain", "K",
                lambda: delete_residue_sidechain_key())

def rotamer_dialog_for_ar():
    with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                               aa_ins_code, aa_atom_name, aa_alt_conf]:
        coot.show_rotamers_dialog(aa_imol, aa_chain_id, aa_res_no,
                             aa_ins_code, aa_alt_conf)
add_key_binding("Rotamers dialog for Active Residue", "Q",
                lambda: rotamer_dialog_for_ar())

def key_binding_func_5():
    active_atom = active_residue()
    if (not active_atom):
        coot.add_status_bar_text("No active residue")
    else:
        imol      = active_atom[0]
        chain_id  = active_atom[1]
        res_no    = active_atom[2]
        ins_code  = active_atom[3]
        atom_name = active_atom[4]
        alt_conf  = active_atom[5]
        name = get_rotamer_name(imol, chain_id, res_no, ins_code)
        if (not name):
            coot.add_status_bar_text("No Name found")
        else:
            if (name == ""):
                coot.add_status_bar_text("No name for this")
            else:
                coot.add_status_bar_text("Rotamer name: " + name)
add_key_binding("Rotamer name in Status Bar", "~", lambda: key_binding_func_5())

refine_residue_sphere_radius_key = 4.5  # Angstroms
add_key_binding("Refine residue in a sphere", "R",
                lambda: fitting.sphere_refine(refine_residue_sphere_radius_key))

def key_binding_func_21():
    if not coot_utils.valid_map_molecule_qm(coot.imol_refinement_map()):
        coot.info_dialog("Must set the refinement map")
    else:
        # not using active atom
        active_atom = active_residue()
        if (not active_atom):
            coot.add_status_bar_text("No active residue")
        else:
            imol      = active_atom[0]
            chain_id  = active_atom[1]
            res_no    = active_atom[2]
            ins_code  = active_atom[3]
            atom_name = active_atom[4]
            alt_conf  = active_atom[5]

            rc_spec = [chain_id, res_no, ins_code]
            ls = coot.residues_near_residue_py(imol, rc_spec, 1.9)
            coot_utils.with_auto_accept([refine_residues, imol, [rc_spec] + ls])

add_key_binding("Neighbours Refine", "h", lambda: key_binding_func_21())

add_key_binding("Regularize Residues in sphere", "B",
                lambda: fitting.sphere_regularize(refine_residue_sphere_radius_key))

def edit_chi_angles_key_func():
    with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                               aa_ins_code, aa_atom_name, aa_alt_conf]:
        coot.edit_chi_angles(aa_imol, aa_chain_id, aa_res_no,
                        aa_ins_code, aa_alt_conf)
        coot.set_moving_atom_move_chis()
        coot.set_graphics_edit_current_chi(1)
add_key_binding("Edit Chi Angles", "X", lambda: edit_chi_angles_key_func())

# BL says:: I like to keep edit chi, so use "Y" for something potentially
# more useful (according to Paule)
add_key_binding("Just One or Next Map", "Y", lambda: coot_utils.just_one_or_next_map())

def jiggle_fit_residue_key():
    with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                               aa_ins_code, aa_atom_name, aa_alt_conf]:
        coot.fit_to_map_by_random_jiggle(aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, 100, 1.0)
add_key_binding("Jiggle Fit Residue", "J", lambda: jiggle_fit_residue_key())

def step_scrollable_map_number():
    maps = coot_utils.map_molecule_list()
    current = coot.scroll_wheel_map()
    if maps:
        l = maps.index(current)
        if (l == len(maps) - 1):
            new = maps[0]
        else:
            new = maps[l+1]
        coot.set_scroll_wheel_map(new)
add_key_binding("Step scrollable map number", "M",
                lambda: step_scrollable_map_number())

def delete_residue_hydrogens_key():
    with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                               aa_ins_code, aa_atom_name, aa_alt_conf]:
        coot.delete_residue_hydrogens(aa_imol, aa_chain_id, aa_res_no,
                                 aa_ins_code, aa_alt_conf)
add_key_binding("Delete Residue Hydrogens", "P",
                lambda: delete_residue_hydrogens_key())

def key_binding_func_9():
    with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                               aa_ins_code, aa_atom_name, aa_alt_conf]:
        coot.additional_representation_by_attributes(aa_imol, aa_chain_id,
                                                aa_res_no, aa_res_no, aa_ins_code,
                                                2, 2, 0.15, 1)
add_key_binding("ball-and-stickify residue", "$", lambda: key_binding_func_9())

add_key_binding("Undo Symmetry View", "V", lambda: coot.undo_symmetry_view())

add_key_binding("Accept Baton Position", "A", lambda: coot.accept_baton_position())

add_key_binding("Cootilus here", "N", lambda: coot.find_nucleic_acids_local(6.0))

def jed_flip_key_func(dir):
    status = coot.jed_flip_intermediate_atoms()
    if status == 0:
       active_atom = active_residue()
       if (not active_atom):
           print("No active atom")
       else:
           imol      = active_atom[0]
           chain_id  = active_atom[1]
           res_no    = active_atom[2]
           ins_code  = active_atom[3]
           atom_name = active_atom[4]
           alt_conf  = active_atom[5]
           coot.jed_flip(imol, chain_id, res_no, ins_code, atom_name, alt_conf, dir)

add_key_binding("JED-Flip", "F", lambda: jed_flip_key_func(0))

add_key_binding("Reverse JED-Flip", "G", lambda: jed_flip_key_func(1))

# Paul's not sure about this one. I likey!
# coot_utils.add_key_binding("Delete this water", "D", lambda: coot.delete_atom(*active_residue()))





import sys
import os

def write_lines(lines, file_name):

    f = open(file_name, "w")
    try:
        for l in lines:
            f.write(l)
    except TypeError as e:
        print(e, "with l", l)
    f.close()

def make_copyright_line(person, copyright_year):

    pp = person
    if person == "Martin": pp = "Martin Noble, University of Oxford"
    if person == "Kevin":  pp = "Kevin Cowtan & University York"

    line = " * Copyright " + copyright_year + " by " + pp + "\n"
    return line

def insert_new_header(person, file_name, copyright_year, lines):

    copyright_line = make_copyright_line(person, copyright_year)
    file_line = " * " + file_name + "\n"
    author_line = " * Author: " + person + "\n"
    if person == "Martin": author_line = " * Author: Martin Noble\n"

    new_lines = [ "/*\n",
                  file_line,
                  " *\n",
                  copyright_line,
                  author_line,
                  " *\n",
                  " * This file is part of Coot\n",
                  " *\n",
                  " * This program is free software; you can redistribute it and/or modify\n",
                  " * it under the terms of the GNU Lesser General Public License as published\n",
                  " * by the Free Software Foundation; either version 3 of the License, or (at\n",
                  " * your option) any later version.\n",
                  " *\n",
                  " * This program is distributed in the hope that it will be useful, but\n",
                  " * WITHOUT ANY WARRANTY; without even the implied warranty of\n",
                  " * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU\n",
                  " * General Public License for more details.\n",
                  " *\n",
                  " * You should have received a copy of the GNU General Public License\n",
                  " * along with this program; if not, write to the Free Software\n",
                  " * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA\n",
                  " * 02110-1301, USA\n",
                  " */\n" ]

    el = enumerate(new_lines)
    for idx,new_line in el:
        lines.insert(idx, new_line)

def get_copyright_year(line):

    y = ""
    parts = line.split()
    # now find the first part after "Copyright" that contains "20"
    idx_copyright = -1
    for idx,part in enumerate(parts):
        if part == "Copyright":
            idx_copyright = idx
        if part == "(C)":
            idx_copyright = idx
    if idx_copyright != -1:
        for idx,part in enumerate(parts):
            if idx > idx_copyright:
                if "20" in part:
                    y = part
    if "Created on" in line: y = parts[-1]

    return y

def get_author(line):
    parts = line.split()
    author = parts[-2] + " " + parts[-1]
    return author

def replace_lines(file_name, lines):

    done = False

    # ---------------- Martin's code style

    if True:
        raw_top_do_it = False
        if '#include ' in lines[0]: raw_top_do_it = True
        if '#define '  in lines[1]: raw_top_do_it = True
        if '#ifndef '  in lines[1]: raw_top_do_it = True
        if '#ifndef '  in lines[0]: raw_top_do_it = True

        if raw_top_do_it:
            insert_new_header("Martin", file_name, "2009", lines)
            done = True

        for idx,line in enumerate(lines):
            if done: continue
            if idx < 10:
                if line == " */\n":
                    for i in range(idx+1):
                        lines.pop(0)
                    insert_new_header("Martin", file_name, "2009", lines)
                    done = True

        fparts = 0
        for idx,line in enumerate(lines):
            if done: continue
            if idx > 10: continue
            if "//  MoleculesToTriangles"    in line: fparts += 1
            if "//  Created by Martin Noble" in line: fparts += 1
            if "//  Copyright "              in line: fparts += 1
            if "Copyright" in line: copyright_year = get_copyright_year(line)
            if line == "//\n":
                if fparts == 3:
                    for i in range (idx+1):
                        lines.pop(0)
                    insert_new_header("Martin", file_name, copyright_year, lines)
                    done = True

    # ---------------- Kevin's code style

    if False:
        copyright_year = ""
        for idx,line in enumerate(lines):
            if done: continue
            if " (C) "        in line: copyright_year = get_copyright_year(line)
            if " Copyright "  in line: copyright_year = get_copyright_year(line)
            if " Created on:" in line: copyright_year = get_copyright_year(line)
            if " Author: "    in line: author = get_author(line)
            if "York all rights reserved" in line:
                lines.pop(idx)
                lines.pop(idx-1)
                insert_new_header("Kevin Cowtan", file_name, copyright_year, lines)
                done = True

            if idx < 10:
                if " */\n" in line:
                    if len(line) == 4:
                        for i in range(idx+1):
                            lines.pop(0)
                        insert_new_header(author, file_name, copyright_year, lines)
                        done = True

if len(sys.argv) > 1:
    fn = sys.argv[1]
    new_file_name = fn + ".new"
    print("Converting header:", fn)
    f = open(fn)
    lines = f.readlines()
    if len(lines) > 6:
        replace_lines(fn, lines)
        write_lines(lines, new_file_name)



#!/usr/bin/env python
"""This module is a simple wrapper for RCrane modules to make RCrane easy to import and to avoid polluting Coot's Python namespace"""

# Copyright 2010, 2011, 2012 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from traceGui import TraceGui as newTrace
from about import createAboutDialog
from rotamerize import pickAtoms as newRotamerize
from rotamerize import storeMenuItem as storeRotamerizeMenuItem
from rotamerize import setRotamerizeMaxNucleotides, getRotamerizeMaxNucleotides
from citationPopup import dontShowPopup as dontShowCitationPopup
from calcCoords import disablePhenixRestraints, enablePhenixRestraints, usingPhenixRestraints
from menu import createRCraneMenu
from preLaunchCheck import checkCootAndReportErrors
from extendChain import pickAtoms as newExtendChain
from extendChain import storeMenuItem as storeExtendChainMenuItem

try:
    from debug import enableDebugging
except ImportError:
    #if we can't find the debug module, then just replace the enableDebugging function with a stub
    #(the debug module isn't included in the distributable)
    def enableDebugging(): return False

newTrace5to3 = lambda: newTrace(direction = 3)
    #a convenience function for users who are setting keyboard shortcuts
newTrace3to5 = lambda: newTrace(direction = 5)
    #a convenience function to allow for consistant naming between different trace directions
newRotamerizeWithoutDensity = lambda: newRotamerize(ignoreDensity = True)
    #a convenience function to rotamerize without density
    
def createRCraneMenuWithCheck():
    """If Coot passes the RCrane pre-launch checks, then create the RCrane menu.
    
    ARGUMENTS:
        None
    RETURNS:
        True if Coot passed the RCrane pre-launch checks
        False otherwise
    """
    
    if checkCootAndReportErrors():
        createRCraneMenu()
        return True
    else:
        return False



# 04_cootaneer.py
# Copyright 2007, 2008 by The University of York
# Author: Bernhard Lohkamp
# Copyright 2007, 2008 by The University of Oxford
# Author: Paul Emsley

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA

import unittest
import os
import coot
import coot_utils
import begin

rnase_pir     = os.path.join(begin.unittest_data_dir, "rnase.pir")
poly_ala_frag = os.path.join(begin.unittest_data_dir, "crashes_on_cootaneering-v2.pdb")

class CootaneerTestFunctions(unittest.TestCase):

    def test01_0(self):
        """Assignment of new PIR sequence overwrites old assignment"""

        imol = begin.unittest_pdb("tutorial-modern.pdb")
        seq_1 = "ACDEFGHIKLMNPQ"
        seq_2 = "ACDEFGHIKLMNPQRST"
        pir_seq_1 = ">test\n\n" + seq_1 + "*"
        pir_seq_2 = ">test\n\n" + seq_2 + "*"

        coot.assign_pir_sequence(imol, "A", pir_seq_1)
        si = coot.sequence_info(imol)
        chains = [item[0] for item in si]
        self.assertTrue(chains.count("A") > 0,
                        """chain A not found in sequence info\n
                        Bad sequence assignment""")
        seq = si[chains.index("A")][1]

        # slightly different logic to greg test, since I cannot
        # query list as dict in python
        print("debug seq", seq)

        self.assertTrue(seq == seq_1,
                        "bad sequence - not matched %s vs %s with 'assoc'"
                        %(seq, seq_1))

        coot.assign_pir_sequence(imol, "A", pir_seq_2)
        si = coot.sequence_info(imol)
        chains = [item[0] for item in si]
        self.assertTrue(chains.count("A") > 0,
                        """chain A not found in sequence info\n
                        Bad sequence assignment - 2""")
        seq = si[chains.index("A")][1]

        self.assertTrue(seq == seq_2,
                        "bad sequence - not matched %s vs %s with 'assoc'"
                        %(seq, seq_2))


    def test02_0(self):
        """Cootaneer Beta Strand"""
        imol_model = coot.read_pdb(poly_ala_frag)
        imol_map = coot.make_and_draw_map(begin.rnase_mtz(), "FWT", "PHWT", "", 0, 0)

        self.assertTrue(coot_utils.valid_model_molecule_qm(imol_model),
                        "bad imol_model: %i from file %s" %(imol_model, poly_ala_frag))
        self.assertTrue(os.path.isfile(rnase_pir), "missing rnase pir file")

        seq_text = coot_utils.file_to_string(rnase_pir)
        #assign_sequence_from_file(imol_model, rnase_pir)
        coot.assign_pir_sequence(imol_model, "A", seq_text)

        coot.set_rotation_centre(64.271, 7.036, 14.42)

        n_atom = coot_utils.closest_atom(imol_model)
        self.assertTrue(n_atom, "missing closest atom")
        imol     = n_atom[0]
        chain_id = n_atom[1]
        resno    = n_atom[2]
        inscode  = n_atom[3]
        at_name  = n_atom[4]
        alt_conf = n_atom[5]
        print("   Cootaneering: imol %s chain-id %s resno %s inscode %s at-name %s alt-conf %s" \
              %(imol, chain_id, resno, inscode, at_name, alt_conf))

        coot.cootaneer(imol_map, imol, [chain_id, resno, inscode, at_name, alt_conf])

